#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

import os
import json
from pyomo.common.fileutils import this_file_dir


def load_generic_ccs_costing_dictionary(path=None):
    """
    Custom dictionaries have been added as a way to add new scaling equations
    that are not based on the Bituminous Baseline report.
    New cost accounts include:

        - 5.1.a.epri can be used to cost the Cansolv carbon capture system
          between 90% and 97.7% capture rate. It is only valid for a flue gas CO2
          concentration of 4%. The additional cost component of the CO2 system is
          in account 5.1.b and remains unchanged.

        - 6.1 can be used to cost the MEA solvent-based carbon capture system,
          using the reference: Kangkang Li, Ashleigh Cousins, Hai You, Paul Feron,
          Weilang Luo, Jian Chen (2016). Systematic study of aqueous
          monoethanolamine (MEA)-based CO2 capture process: Techno-economic
          assessment of the MEA process and its improvements. Applied Energy, 165,
          648-659.
    """
    if path is None:
        directory = this_file_dir()
    else:
        directory = path

    generic_ccs_costing_exponents = {
        "6": {
            "5.1.a.epri": {
                "Account Name": "Cansolv CO2 Removal System",
                "Exponent": 2.788,
                "Process Parameter": "CO2 Flowrate",
            },
            "6.1.ccs": {
                "Account Name": "MEA solvent capture system absorber",
                "Exponent": 0.6,
                "Process Parameter": "Absorber volume",
            },
            "6.2.ccs": {
                "Account Name": "MEA solvent capture system absorber packing",
                "Exponent": 0.6,
                "Process Parameter": "Absorber packing volume",
            },
            "6.3.ccs": {
                "Account Name": "MEA solvent capture system stripper",
                "Exponent": 0.6,
                "Process Parameter": "Stripper volume",
            },
            "6.4.ccs": {
                "Account Name": "MEA solvent capture system stripper packing",
                "Exponent": 0.6,
                "Process Parameter": "Stripper packing volume",
            },
            "6.5.ccs": {
                "Account Name": "MEA solvent capture system stripper condenser",
                "Exponent": 0.6,
                "Process Parameter": "Stripper condenser area",
            },
            "6.6.ccs": {
                "Account Name": "MEA solvent capture system stripper reboiler",
                "Exponent": 0.6,
                "Process Parameter": "Stripper reboiler area",
            },
            "6.7.ccs": {
                "Account Name": "MEA solvent capture system lean rich heat exchanger",
                "Exponent": 0.6,
                "Process Parameter": "Lean rich heat exchanger area",
            },
            "6.8.ccs": {
                "Account Name": "MEA solvent capture system lean solvent cooler",
                "Exponent": 0.6,
                "Process Parameter": "Lean solvent cooler area",
            },
            "6.9.1.ccs": {
                "Account Name": "MEA solvent capture system flue gas blower cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.9.2.ccs": {
                "Account Name": "MEA solvent capture system flue gas blower cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.10.1.ccs": {
                "Account Name": "MEA solvent capture system flue gas direct contact cooler cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.10.2.ccs": {
                "Account Name": "MEA solvent capture system flue gas direct contact cooler cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.11.1.ccs": {
                "Account Name": "MEA solvent capture system flue gas direct contact cooler packing cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.11.2.ccs": {
                "Account Name": "MEA solvent capture system flue gas direct contact cooler packing cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.12.1.ccs": {
                "Account Name": "MEA solvent capture system pretreatment pump cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.12.2.ccs": {
                "Account Name": "MEA solvent capture system pretreatment pump cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.13.1.ccs": {
                "Account Name": "MEA solvent capture system pretreatment cooler cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.13.2.ccs": {
                "Account Name": "MEA solvent capture system pretreatment cooler cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.14.1.ccs": {
                "Account Name": "MEA solvent capture system pretreatment tank cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.14.2.ccs": {
                "Account Name": "MEA solvent capture system pretreatment tank cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.15.1.ccs": {
                "Account Name": "MEA solvent capture system washing column cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.15.2.ccs": {
                "Account Name": "MEA solvent capture system washing column cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.16.1.ccs": {
                "Account Name": "MEA solvent capture system washing column packing cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.16.2.ccs": {
                "Account Name": "MEA solvent capture system washing column packing cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.17.1.ccs": {
                "Account Name": "MEA solvent capture system washing solvent cooler cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.17.2.ccs": {
                "Account Name": "MEA solvent capture system washing solvent cooler cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.18.1.ccs": {
                "Account Name": "MEA solvent capture system washing solvent pump cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.18.2.ccs": {
                "Account Name": "MEA solvent capture system washing solvent pump cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.19.1.ccs": {
                "Account Name": "MEA solvent capture system condenser pump cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.19.2.ccs": {
                "Account Name": "MEA solvent capture system condenser pump cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.20.1.ccs": {
                "Account Name": "MEA solvent capture system stripper reflux drum cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.20.2.ccs": {
                "Account Name": "MEA solvent capture system stripper reflux drum cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.21.1.ccs": {
                "Account Name": "MEA solvent capture system lean solvent pump cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.21.2.ccs": {
                "Account Name": "MEA solvent capture system lean solvent pump cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.22.1.ccs": {
                "Account Name": "MEA solvent capture system solvent storage tank cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.22.2.ccs": {
                "Account Name": "MEA solvent capture system solvent storage tank cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.23.1.ccs": {
                "Account Name": "MEA solvent capture system washing solvent tank cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.23.2.ccs": {
                "Account Name": "MEA solvent capture system washing solvent tank cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.24.1.ccs": {
                "Account Name": "MEA solvent capture system solvent stripper reclaimer cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.24.2.ccs": {
                "Account Name": "MEA solvent capture system solvent stripper reclaimer cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.25.1.ccs": {
                "Account Name": "MEA solvent capture system solvent reclaimer cooler cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.25.2.ccs": {
                "Account Name": "MEA solvent capture system solvent reclaimer cooler cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
            "6.26.1.ccs": {
                "Account Name": "MEA solvent capture system solvent filtration cost component 1",
                "Exponent": 0.6,
                "Process Parameter": "CO2 product flow rate",
            },
            "6.26.2.ccs": {
                "Account Name": "MEA solvent capture system solvent filtration cost component 2",
                "Exponent": 0.6,
                "Process Parameter": "Flue gas inlet to absorber",
            },
        }
    }

    generic_ccs_costing_params = {
        "6": {
            "B": {
                "5.1.a.epri": {
                    "BEC": 224191.4,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.18,
                    "Project Contingency": 0.2,
                    "RP Value": 493587.88,
                    "Units": "lb/hr",
                },
                "6.1.ccs": {
                    "BEC": 8376000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 1074.424688,
                    "Units": "m**3",
                },
                "6.2.ccs": {
                    "BEC": 6117000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 791.6813487,
                    "Units": "m**3",
                },
                "6.3.ccs": {
                    "BEC": 2273000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 419.6971436,
                    "Units": "m**3",
                },
                "6.4.ccs": {
                    "BEC": 1464000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 309.2505268,
                    "Units": "m**3",
                },
                "6.5.ccs": {
                    "BEC": 305000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 800,
                    "Units": "m**2",
                },
                "6.6.ccs": {
                    "BEC": 3183000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 4250,
                    "Units": "m**2",
                },
                "6.7.ccs": {
                    "BEC": 1502000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 9100,
                    "Units": "m**2",
                },
                "6.8.ccs": {
                    "BEC": 522000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 1200,
                    "Units": "m**2",
                },
                "6.9.1.ccs": {
                    "BEC": 0.6 * 812000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.9.2.ccs": {
                    "BEC": 0.4 * 812000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.10.1.ccs": {
                    "BEC": 0.6 * 2875000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.10.2.ccs": {
                    "BEC": 0.4 * 2875000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.11.1.ccs": {
                    "BEC": 0.6 * 1883000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.11.2.ccs": {
                    "BEC": 0.4 * 1883000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.12.1.ccs": {
                    "BEC": 0.6 * 108000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.12.2.ccs": {
                    "BEC": 0.4 * 108000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.13.1.ccs": {
                    "BEC": 0.6 * 202000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.13.2.ccs": {
                    "BEC": 0.4 * 202000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.14.1.ccs": {
                    "BEC": 0.6 * 93000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.14.2.ccs": {
                    "BEC": 0.4 * 93000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.15.1.ccs": {
                    "BEC": 0.6 * 2842000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.15.2.ccs": {
                    "BEC": 0.4 * 2842000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.16.1.ccs": {
                    "BEC": 0.6 * 2060000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.16.2.ccs": {
                    "BEC": 0.4 * 2060000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.17.1.ccs": {
                    "BEC": 0.6 * 69000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.17.2.ccs": {
                    "BEC": 0.4 * 69000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.18.1.ccs": {
                    "BEC": 0.6 * 17000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.18.2.ccs": {
                    "BEC": 0.4 * 17000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.19.1.ccs": {
                    "BEC": 0.6 * 35000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.19.2.ccs": {
                    "BEC": 0.4 * 35000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.20.1.ccs": {
                    "BEC": 0.6 * 46000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.20.2.ccs": {
                    "BEC": 0.4 * 46000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.21.1.ccs": {
                    "BEC": 0.6 * 159000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.21.2.ccs": {
                    "BEC": 0.4 * 159000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.22.1.ccs": {
                    "BEC": 0.6 * 342000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.22.2.ccs": {
                    "BEC": 0.4 * 342000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.23.1.ccs": {
                    "BEC": 0.6 * 46000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.23.2.ccs": {
                    "BEC": 0.4 * 46000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.24.1.ccs": {
                    "BEC": 0.6 * 181000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.24.2.ccs": {
                    "BEC": 0.4 * 181000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.25.1.ccs": {
                    "BEC": 0.6 * 180000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.25.2.ccs": {
                    "BEC": 0.4 * 180000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.26.1.ccs": {
                    "BEC": 0.6 * 866000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.26.2.ccs": {
                    "BEC": 0.4 * 866000 / 1e3,
                    "Eng Fee": 0.2,
                    "Process Contingency": 0.12,
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
            }
        }
    }

    # this is a way to "zip together" two generic_ccs_costing dictionaries in the formats above
    # the BEC_units happen to be in thousands of 2013 USD for these accounts
    # for the power plant methods to work, BEC_units must be "$year", "K$year" or
    # "$Myear" for USD, thousands USD and millions USD, respectively

    if not os.path.exists(
        os.path.join(directory, "generic_ccs_costing_data.json")
    ):  # make the dictionary

        generic_ccs_costing_data = generic_ccs_costing_params

        for (
            tech,
            gccsd,
        ) in generic_ccs_costing_data.items():  # do one technology at a time
            for ccs in ["A", "B"]:
                if ccs in gccsd:  # check if CCS = A, for indexing
                    accounts_dict = gccsd[ccs]  # shorter alias
                    for account in accounts_dict.keys():  # do one account at a time
                        if account == "5.1.a.epri":
                            accounts_dict[account][
                                "BEC_units"
                            ] = "K$2018"  # add BEC units as thousands of 2018 USD
                        else:
                            accounts_dict[account][
                                "BEC_units"
                            ] = "K$2013"  # add BEC units as thousands of 2013 USD
                        for accountkey in generic_ccs_costing_exponents[tech][
                            account
                        ].keys():  # get one " exponents"account property at a time
                            accounts_dict[account][accountkey] = (
                                generic_ccs_costing_exponents[tech][account][accountkey]
                            )
                        sorted_accountkeys = sorted(
                            accounts_dict[account]
                        )  # now, sort the accountkeys alphabetically within each account
                        accounts_dict[account] = {
                            key: accounts_dict[account][key]
                            for key in sorted_accountkeys
                        }  # re-add the keys in alphabetical order
                    gccsd[ccs] = (
                        accounts_dict  # use the alias to update the original dictionary
                    )

        with open(
            os.path.join(directory, "generic_ccs_costing_data.json"), "w"
        ) as outfile:
            json.dump(generic_ccs_costing_data, outfile)
        print("Success! New costing dictionary generated.")

    # assuming the dictionary exists, load it so it is importable when called
    with open(os.path.join(directory, "generic_ccs_costing_data.json"), "r") as file:
        generic_ccs_costing_params = json.load(file)
    return generic_ccs_costing_params



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring

# TODO: protected access issues
# pylint: disable=protected-access

import os
import hashlib
from platform import machine
import tarfile
import urllib

from pyomo.common.download import FileDownloader

import idaes
import idaes.logger as idaeslog
from idaes.config import extra_binaries

_log = idaeslog.getLogger(__name__)
_release_base_url = idaes.config.release_base_url


class UnsupportedPlatformError(RuntimeError):
    pass


def hash_file_sha256(fname):
    """Calculate sha256 hash of potentially large files.

    Args:
        fname: path to file to hash

    Returns:
        hash as a string
    """
    with open(fname, "rb") as f:
        fh = hashlib.sha256()
        while True:
            fb = f.read(10000)
            if len(fb) <= 0:
                break
            fh.update(fb)
    return str(fh.hexdigest())


_hash = hash_file_sha256


def _get_file_downloader(insecure, cacert):
    fd = FileDownloader(insecure=insecure, cacert=cacert)
    arch = fd.get_sysinfo()
    return fd, arch


def _get_arch_and_platform(fd, platform):
    arch = fd.get_sysinfo()

    if platform == "auto":
        platform = arch[0]
    if platform == "linux":
        platform = fd.get_os_version().replace(".", "")
        _log.debug(f"Detected Linux distribution: {platform}")

    return arch, platform


def _get_release_platform(platform):
    # Check if platform (OS) maps to another platform
    platform = idaes.config.canonical_distro(platform)

    # Get machine type (e.g. x86_64, ...)
    mach = idaes.config.canonical_arch(machine())

    # full platform name
    platform = f"{platform}-{mach}"
    # See if machine is supported
    if platform not in idaes.config.base_platforms:
        raise UnsupportedPlatformError(f"Unsupported platform: {platform}.")
    _log.debug(f"Downloading binaries for {platform}")

    return platform


def _get_release_url(release, url):
    if release is not None:
        url = "/".join([_release_base_url, release])
    else:
        _log.debug("Release not specified.")

    if url is None:
        _log.debug("No release or URL was provided.")
        # PYLINT-TODO
        # pylint: disable-next=broad-exception-raised
        raise Exception("Must provide a location to download binaries")

    if url.endswith("/"):
        url = url[0:-1]  # if url ends with "/" remove it for proper join later

    _log.debug(f"Downloading binaries from {url}")
    return url


def _get_checksum_paths(to_path, release):
    check_to = os.path.join(to_path, f"sha256sum_{release}.txt")
    check_from = idaes.config.release_checksum_url.format(release)
    _log.debug(f"Getting checksum file {check_from}")

    return check_to, check_from


def _download_checksum(fd, check_to, check_from):
    fd.set_destination_filename(check_to)
    fd.get_binary_file(check_from)


def _read_checksum_file(check_to):
    # read the hashes file and store them in checksum dict
    checksum = {}  # storage for hashes if install from release

    with open(check_to, "r") as f:
        for i in range(1000):  # pylint: disable=unused-variable
            line = f.readline(1000)
            if line == "":
                break
            line = line.split(sep="  ")
            checksum[line[1].strip()] = line[0].strip()

    return checksum


def _get_checksums(fd, to_path, release, nochecksum):
    if nochecksum:
        _log.debug("Skip release checksum verification at user request.")
        return False  # default if not checking checksums

    check_to, check_from = _get_checksum_paths(to_path, release)
    _download_checksum(fd, check_to, check_from)

    return _read_checksum_file(check_to)


def _create_download_package(platform, to_path, url, extra, extras_only, library_only):
    pname = []
    ptar = []
    ftar = []
    furl = []

    def _add_pack(name):
        f = f"idaes-{name}-{platform}.tar.gz"
        ftar.append(f)
        ptar.append(os.path.join(to_path, f))
        pname.append(name)
        furl.append("/".join([url, f]))

    for e in extra:
        if e not in extra_binaries:
            _log.warning(f"Unknown extra package {e}, not installed.")
            continue
        if platform not in extra_binaries[e]:
            _log.warning(
                f"Extra package {e} not available for {platform}, not installed."
            )
            continue
        _add_pack(e)  # you have to explicitly ask for extras so assume you want
    if not extras_only:
        _add_pack("lib")
    if not library_only and not extras_only:
        _add_pack("solvers")

    return pname, ptar, ftar, furl


def _download_package(fd, name, frm, to, platform):
    _log.debug(f"Getting {name} from: {frm}")
    _log.debug(f"Saving solvers to: {to}")
    fd.set_destination_filename(to)

    try:
        fd.get_binary_file(frm)
    except urllib.error.HTTPError:
        # PYLINT-TODO
        # pylint: disable-next=broad-exception-raised
        raise Exception(f"{name} binaries are unavailable for {platform}")


def _verify_checksums(checksum, pname, ptar, ftar):
    # If release checksum is not False, nochecksum opt allows hash to be ignored
    if checksum:
        for n, p, f in zip(pname, ptar, ftar):
            hash_l = _hash(p)
            _log.debug(f"{n} Hash {hash_l}")
            if checksum.get(f, "") != hash_l:
                # PYLINT-TODO
                # pylint: disable-next=broad-exception-raised
                raise Exception(f"{n} hash does not match expected")


def _splitpath(path):
    """Split path into elements

    This routine is similar to str.split(), except when the pieces are
    re-joined by os.path.join(), the result will still be an absolute
    path (if ``path`` was an absolute path).  Note that the path
    elements are returned in reverse order (which is convenient for
    _resolve_path below)

    """
    ans = []
    while 1:
        a, b = os.path.split(path)
        if b:
            ans.append(b)
        if a == path:
            if a:
                ans.append(a)
            return ans
        path = a


def _resolve_path(path, links):
    """Resolve path to the actual filesystem location

    This routine is similar to os.realpath(), but with the added feature
    of recognizing not only hard / symbolic links on the filesystem, but
    also links in the ``links`` dict that have not yet been committed to
    the filesystem.

    """
    dirs = _splitpath(path)
    path = ""
    target = None
    while dirs:
        path = os.path.realpath(os.path.join(path, dirs.pop()))
        if path in links:
            target = path
        while target in links:
            if links[target][0] == tarfile.SYMTYPE:
                if os.path.isabs(links[target][1]):
                    path = ""
                else:
                    path = os.path.dirname(path)
                dirs.extend(_splitpath(links[target][1]))
                target = None
            else:  # links[target][0] == tarfile.LNKTYPE:
                target_dir, target_name = os.path.split(links[target][1])
                target = os.path.join(_resolve_path(target_dir, links), target_name)
        if target is not None:
            path = target
            target = None
    return path


def _verify_tar_member_targets(tar, to_path, links=None):
    """Check tar for unsafe members (references outside to_path)

    Nominally, this would be a simple task: verify that every member of
    the tarfile (when converted to an absolute path and normalized)
    starts with the to_path.  Unfortunately, symbolic links make this
    more difficult.  It is not sufficient to just check the absolute
    path of the link target, as sequences of links could walk out of the
    target space.  In this routine, we attempt to identify symbolic
    links and resolve the actual location where each tar member would be
    stored on the filesystem.

    Example::

        [DIRTYPE] dir1
        [SYMTYPE] dir1/dir2 -> ..
        [SYMTYPE] dir1/dir2/dir3 -> ..
        [REGTYPE] dir1/dir2/dir3/file  # <-- this is ../file

    Parameters
    ----------
    tar : TarFile
        tarfile to verify

    to_path : str
        target directory to extract tarfile into

    links : dict[str, tuple[bytes, str]]
        dictionary that records "virtual" symbolic links; that is,
        symbolic links in the tarfile that have not yet been extracted,
        but will exist during / after the extraction process.  This maps
        the actual symbolic link location to a (type, target) tuple.
        The target is a resolved (absolute) path for hard links and a
        potentially relative path for soft links.

    """
    to_path = os.path.realpath(to_path)

    if links is None:
        # A dict mapping {from: (type, to)}
        links = {}

    for member in tar.getmembers():
        member_path = _resolve_path(os.path.join(to_path, member.name), links)
        if not member_path.startswith(to_path):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                f"Tarball {tar.name} contained potentially unsafe member "
                f"{member.name} that would extract to {member_path} outside "
                f"target directory {to_path}"
            )
        if member.issym():
            links[member_path] = (member.type, member.linkname)
        elif member.islnk() and member.name != member.linkname:
            links[member_path] = (member.type, os.path.join(to_path, member.linkname))


def download_binaries(
    release=None,
    url=None,
    insecure=False,
    cacert=None,
    verbose=False,
    platform="auto",
    nochecksum=False,
    library_only=False,
    no_download=False,
    extras_only=False,
    extra=(),
    to_path=None,
    alt_path=None,
):
    """
    Download IDAES solvers and libraries and put them in the right location. Need
    to supply either local or url argument.

    Args:
        url (str): a url to download binary files to install files
        to_path: for testing only, an alternate subdirectory of the idaes data
                 directory for a test binary install.
    Returns:
        None
    """
    if verbose:
        _log.setLevel(idaeslog.DEBUG)
    if no_download:
        nochecksum = True

    # set the locations to download files to, to_path is an alternate
    # subdirectory of idaes.data_directory that can optionally be used to test
    # this function without interfering with anything else.  It's a subdirectory
    # of the data directory because that should be a safe place to store some
    # test files.
    if alt_path is not None:
        to_path = os.path.abspath(alt_path)
    if to_path is None:
        to_path = idaes.bin_directory
    else:
        to_path = os.path.join(idaes.data_directory, to_path)
    idaes._create_bin_dir(to_path)

    # Create FileDownloader object
    fd = FileDownloader(insecure=insecure, cacert=cacert)

    # Get platform information
    arch, platform = _get_arch_and_platform(  # pylint: disable=unused-variable
        fd, platform
    )
    platform = _get_release_platform(platform)

    # Get release url
    url = _get_release_url(release, url)

    # Get checksum
    checksum = _get_checksums(fd, to_path, release, nochecksum)

    # Set the binary file destinations
    pname, ptar, ftar, furl = _create_download_package(
        platform, to_path, url, extra, extras_only, library_only
    )

    if no_download:
        d = {
            "release": release,
            "platform": platform,
        }
        for n, p, u in zip(pname, ptar, furl):
            d[u] = p
        return d

    # download packages
    for n, p, u in zip(pname, ptar, furl):
        _download_package(fd, n, frm=u, to=p, platform=platform)

    # Verify checksums
    _verify_checksums(checksum, pname, ptar, ftar)

    # Extract solvers
    links = {}
    for n, p in zip(pname, ptar):
        _log.debug(f"Extracting files in {p} to {to_path}")
        with tarfile.open(p, "r") as f:
            _verify_tar_member_targets(f, to_path, links)
            f.extractall(to_path)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES Component objects

@author: alee
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import Set, Param, Var, units as pyunits
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf, Bool

from idaes.core.base.process_base import declare_process_block_class, ProcessBlockData
from idaes.core.base.phases import PhaseType
from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError
from idaes.core.util.misc import set_param_from_config
import idaes.logger as idaeslog


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Component")
class ComponentData(ProcessBlockData):
    """
    Standard component object class.
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "valid_phase_types",
        ConfigValue(
            domain=ListOf(PhaseType),
            doc="List of valid PhaseTypes (Enums) for this Component.",
        ),
    )

    CONFIG.declare(
        "elemental_composition",
        ConfigValue(
            domain=dict,
            description="Elemental composition of component",
            doc="Dict containing elemental composition in the form element "
            ": stoichiometry",
        ),
    )

    CONFIG.declare(
        "henry_component",
        ConfigValue(
            domain=dict,
            description="Phases in which component follows Henry's Law",
            doc="Dict indicating phases in which component follows Henry's "
            "Law (keys) with values indicating form of law.",
        ),
    )

    CONFIG.declare(
        "vol_mol_liq_comp",
        ConfigValue(
            description="Method to use to calculate liquid phase molar volume",
            doc="Method to use to calculate liquid phase molar volume. Users "
            "need only define either vol_mol_liq_comp or dens_mol_liq_comp.",
        ),
    )
    CONFIG.declare(
        "vol_mol_sol_comp",
        ConfigValue(
            description="Method to use to calculate solid phase molar volume",
            doc="Method to use to calculate solid phase molar volume. Users "
            "need only define either vol_mol_sol_comp or dens_mol_sol_comp.",
        ),
    )
    CONFIG.declare(
        "dens_mol_liq_comp",
        ConfigValue(
            description="Method to use to calculate liquid phase molar density",
            doc="Method to use to calculate liquid phase molar density. Users "
            "need only define either vol_mol_liq_comp or dens_mol_liq_comp.",
        ),
    )
    CONFIG.declare(
        "dens_mol_sol_comp",
        ConfigValue(
            description="Method to use to calculate solid phase molar density",
            doc="Method to use to calculate solid phase molar density. Users "
            "need only define either vol_mol_sol_comp or dens_mol_sol_comp.",
        ),
    )

    CONFIG.declare(
        "cp_mol_liq_comp",
        ConfigValue(description="Method to calculate liquid component specific heats"),
    )
    CONFIG.declare(
        "cp_mol_sol_comp",
        ConfigValue(description="Method to calculate solid component specific heats"),
    )
    CONFIG.declare(
        "cp_mol_ig_comp",
        ConfigValue(
            description="Method to calculate ideal gas component specific heats"
        ),
    )
    CONFIG.declare(
        "enth_mol_liq_comp",
        ConfigValue(
            description="Method to calculate liquid component molar enthalpies"
        ),
    )
    CONFIG.declare(
        "enth_mol_sol_comp",
        ConfigValue(description="Method to calculate solid component molar enthalpies"),
    )
    CONFIG.declare(
        "enth_mol_ig_comp",
        ConfigValue(
            description="Method to calculate ideal gas component molar enthalpies"
        ),
    )
    CONFIG.declare(
        "entr_mol_liq_comp",
        ConfigValue(description="Method to calculate liquid component molar entropies"),
    )
    CONFIG.declare(
        "entr_mol_sol_comp",
        ConfigValue(description="Method to calculate solid component molar entropies"),
    )
    CONFIG.declare(
        "entr_mol_ig_comp",
        ConfigValue(
            description="Method to calculate ideal gas component molar entropies"
        ),
    )

    CONFIG.declare(
        "diffus_phase_comp",
        ConfigValue(
            description="Method to calculate component diffusivities in each "
            "phase. Must be a dict with keys being phase names."
        ),
    )
    CONFIG.declare(
        "visc_d_phase_comp",
        ConfigValue(
            description="Method to calculate pure component viscosities in each "
            "phase. Must be a dict with keys being phase names."
        ),
    )
    CONFIG.declare(
        "therm_cond_phase_comp",
        ConfigValue(
            description="Method to calculate pure component thermal conductivities in each "
            "phase. Must be a dict with keys being phase names."
        ),
    )
    CONFIG.declare(
        "has_vapor_pressure",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Flag indicating whether component has a vapor pressure",
        ),
    )
    CONFIG.declare(
        "pressure_sat_comp",
        ConfigValue(description="Method to use to calculate saturation pressure"),
    )
    CONFIG.declare(
        "relative_permittivity_liq_comp",
        ConfigValue(
            description="Method to use to calculate liquid phase relative permittivity"
        ),
    )

    CONFIG.declare(
        "phase_equilibrium_form",
        ConfigValue(
            domain=dict,
            description="Form of phase equilibrium constraints for component",
        ),
    )

    CONFIG.declare(
        "parameter_data",
        ConfigValue(
            default={},
            domain=dict,
            description="Dict containing initialization data for parameters",
        ),
    )

    CONFIG.declare(
        "_component_list_exists",
        ConfigValue(
            default=False,
            doc="Internal config argument indicating whether component_list "
            "needs to be populated.",
        ),
    )
    CONFIG.declare(
        "_electrolyte",
        ConfigValue(
            default=False,
            doc="Internal config argument indicating whether electrolyte "
            "component_lists needs to be populated.",
        ),
    )

    def build(self):
        super().build()

        # If the component_list does not exist, add reference to new Component
        # The IF is mostly for backwards compatibility, to allow for old-style
        # property packages where the component_list already exists but we
        # need to add new Component objects

        # WE control the config block, ignore warnings about protected access
        if not self.config._component_list_exists:  # pylint: disable=protected-access
            if not self.config._electrolyte:  # pylint: disable=protected-access
                self._add_to_component_list()
            else:
                self._add_to_electrolyte_component_list()

        base_units = self.parent_block().get_metadata().default_units

        # Create Param for molecular weight if provided
        param_dict = {"mw": base_units.MOLECULAR_WEIGHT}
        for p, u in param_dict.items():
            if p in self.config.parameter_data:
                self.add_component(p, Param(mutable=True, units=u))
                set_param_from_config(self, p)

        # Create Vars for common parameters
        var_dict = {
            "compress_fact_crit": pyunits.dimensionless,
            "dens_mol_crit": base_units.DENSITY_MOLE,
            "omega": pyunits.dimensionless,
            "pressure_crit": base_units.PRESSURE,
            "temperature_crit": base_units.TEMPERATURE,
        }
        for p, u in var_dict.items():
            if p in self.config.parameter_data:
                self.add_component(p, Var(units=u))
                set_param_from_config(self, p)

    def is_solute(self):
        raise TypeError(
            "{} Generic Component objects do not support is_solute() method. "
            "Use a Solvent or Solute Component instead.".format(self.name)
        )

    def is_solvent(self):
        raise TypeError(
            "{} Generic Component objects do not support is_solvent() method. "
            "Use a Solvent or Solute Component instead.".format(self.name)
        )

    def _add_to_component_list(self):
        """
        Method to add reference to new Component in component_list
        """
        parent = self.parent_block()
        try:
            comp_list = getattr(parent, "component_list")
            comp_list.add(self.local_name)
        except AttributeError:
            # Parent does not have a component_list yet, so create one
            parent.component_list = Set(initialize=[self.local_name], ordered=True)

    def _add_to_electrolyte_component_list(self):
        """
        Special case method for adding references to new Component in
        component_lists for electrolyte systems,

        New Component types should overload this method
        """
        parent = self.parent_block()
        parent._non_aqueous_set.add(self.local_name)  # pylint: disable=protected-access

    def _is_phase_valid(self, phase):
        # If no valid phases assigned
        if self.config.valid_phase_types is None:
            try:
                if phase.is_aqueous_phase():
                    # If this is an aqueous phase, check for validaity
                    return self._is_aqueous_phase_valid()
            except AttributeError:
                raise TypeError(
                    "{} Phase {} is not a valid phase object or is undeclared."
                    " Please check your phase declarations.".format(self.name, phase)
                )
            # Otherwise assume all are valid
            return True

        # Check for behaviour of phase, and see if that is a valid behaviour
        # for component.
        elif phase.is_liquid_phase():
            # Check if this is an aqueous phase
            if phase.is_aqueous_phase():
                if (
                    self._is_aqueous_phase_valid()
                    and PhaseType.aqueousPhase in self.config.valid_phase_types
                ):
                    return True
                else:
                    return False
            elif PhaseType.liquidPhase in self.config.valid_phase_types:
                return True
            else:
                return False
        elif (
            phase.is_vapor_phase()
            and PhaseType.vaporPhase in self.config.valid_phase_types
        ):
            return True
        elif (
            phase.is_solid_phase()
            and PhaseType.solidPhase in self.config.valid_phase_types
        ):
            return True
        else:
            return False

    def _is_aqueous_phase_valid(self):
        # Method to indicate if a component type is stable in the aqueous phase
        # General components may not appear in aqueous phases
        return False


# TODO : What about LLE systems where a species is a solvent in one liquid
# phase, but a solute in another?
@declare_process_block_class("Solute", block_class=Component)
class SoluteData(ComponentData):
    """
    Component type for species which should be considered as solutes in
    LiquidPhases.
    """

    def is_solute(self):
        return True

    def is_solvent(self):
        return False

    def _is_aqueous_phase_valid(self):
        return True

    def _add_to_component_list(self):
        """
        Method to add reference to new Component in component_list
        """
        super()._add_to_component_list()

        # Also add to solute_set
        parent = self.parent_block()
        try:
            comp_list = getattr(parent, "solute_set")
            comp_list.add(self.local_name)
        except AttributeError:
            # Parent does not have a solute_set yet, so create one
            parent.solute_set = Set(initialize=[self.local_name], ordered=True)

    def _add_to_electrolyte_component_list(self):
        """
        Special case method for adding references to new Component in
        component_lists for electrolyte systems,

        New Component types should overload this method
        """
        parent = self.parent_block()
        parent.solute_set.add(self.local_name)


# TODO : What about LLE systems where a species is a solvent in one liquid
# phase, but a solute in another?
@declare_process_block_class("Solvent", block_class=Component)
class SolventData(ComponentData):
    """
    Component type for species which should be considered as solvents in
    LiquidPhases.
    """

    def is_solute(self):
        return False

    def is_solvent(self):
        return True

    def _is_aqueous_phase_valid(self):
        return True

    def _add_to_component_list(self):
        """
        Method to add reference to new Component in component_list
        """
        super()._add_to_component_list()

        # Also add to solvent_set
        parent = self.parent_block()
        try:
            comp_list = getattr(parent, "solvent_set")
            comp_list.add(self.local_name)
        except AttributeError:
            # Parent does not have a solvent_list yet, so create one
            parent.solvent_set = Set(initialize=[self.local_name], ordered=True)

    def _add_to_electrolyte_component_list(self):
        """
        Special case method for adding references to new Component in
        component_lists for electrolyte systems,

        New Component types should overload this method
        """
        parent = self.parent_block()
        parent.solvent_set.add(self.local_name)


@declare_process_block_class("Ion", block_class=Component)
class IonData(SoluteData):
    """
    Component type for ionic species. These can exist only in AqueousPhases,
    and are always solutes.
    """

    CONFIG = SoluteData.CONFIG()

    # Remove valid_phase_types argument, as ions are aqueous phase only
    del CONFIG["valid_phase_types"]
    # Set as not having a vapor pressure
    has_psat = CONFIG.get("has_vapor_pressure")
    has_psat.set_value(False)
    has_psat.set_default_value(False)
    has_psat.set_domain(In([False]))

    CONFIG.declare("charge", ConfigValue(domain=int, doc="Charge of ionic species."))

    def _is_phase_valid(self, phase):
        return phase.is_aqueous_phase()

    def _is_aqueous_phase_valid(self):
        return True

    def _add_to_component_list(self):
        """
        Ions should not be used outside of electrolyte property methods
        """
        raise PropertyPackageError(
            "{} Ion Component types should only be used with Aqueous "
            "Phases".format(self.name)
        )

    def _add_to_electrolyte_component_list(self):
        """
        Special case method for adding references to new Component in
        component_lists for electrolyte systems,

        New Component types should overload this method
        """
        raise NotImplementedError(
            "{} The IonData component class is intended as a base class for "
            "the AnionData and CationData classes, and should not be used "
            "directly".format(self.name)
        )


@declare_process_block_class("Anion", block_class=Component)
class AnionData(IonData):
    """
    Component type for anionic species. These can exist only in AqueousPhases,
    and are always solutes.
    """

    CONFIG = IonData.CONFIG()

    def build(self):
        super().build()

        # Validate charge config argument
        if self.config.charge is None:
            raise ConfigurationError(
                "{} was not provided with a value for charge.".format(self.name)
            )
        elif self.config.charge >= 0:
            raise ConfigurationError(
                "{} received invalid value for charge configuration argument."
                " Anions must have a negative charge.".format(self.name)
            )

    def _add_to_electrolyte_component_list(self):
        """
        Special case method for adding references to new Component in
        component_lists for electrolyte systems,

        New Component types should overload this method
        """
        parent = self.parent_block()
        parent.anion_set.add(self.local_name)


@declare_process_block_class("Cation", block_class=Component)
class CationData(IonData):
    """
    Component type for cationic species. These can exist only in AqueousPhases,
    and are always solutes.
    """

    CONFIG = IonData.CONFIG()

    def build(self):
        super().build()

        # Validate charge config argument
        if self.config.charge is None:
            raise ConfigurationError(
                "{} was not provided with a value for charge.".format(self.name)
            )
        elif self.config.charge <= 0:
            raise ConfigurationError(
                "{} received invalid value for charge configuration argument."
                " Cations must have a positive charge.".format(self.name)
            )

    def _add_to_electrolyte_component_list(self):
        """
        Special case method for adding references to new Component in
        component_lists for electrolyte systems,

        New Component types should overload this method
        """
        parent = self.parent_block()
        parent.cation_set.add(self.local_name)


@declare_process_block_class("Apparent", block_class=Component)
class ApparentData(SoluteData):
    """
    Component type for apparent species. Apparent species are those compounds
    that are not stable in aqueous phases and immediately dissociate, however
    they may be stable in other phases (e.g. salts).
    """

    CONFIG = SoluteData.CONFIG()
    CONFIG.declare(
        "dissociation_species",
        ConfigValue(
            domain=dict,
            default=None,
            description="Dict of dissociation species",
            doc="Dict of true species that this species will dissociate into "
            "upon dissolution along with stoichiometric coefficients.",
        ),
    )

    def build(self):
        super().build()

        # Make sure dissoication species were set
        if self.config.dissociation_species is None:
            raise ConfigurationError(
                f"{self.name} dissoication_species argument was not set. "
                f"Apparent components require the dissociation species to be "
                f"defined."
            )

    def _is_aqueous_phase_valid(self):
        return True

    def _add_to_electrolyte_component_list(self):
        """
        Special case method for adding references to new Component in
        component_lists for electrolyte systems,

        New Component types should overload this method
        """
        parent = self.parent_block()
        parent._apparent_set.add(self.local_name)  # pylint: disable=protected-access


__all_components__ = [Component, Solute, Solvent, Ion, Anion, Cation, Apparent]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for rate forms
"""

import pytest

from pyomo.environ import ConcreteModel, Var, units as pyunits, value
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
    ConcentrationForm,
)
from idaes.models.properties.modular_properties.reactions.rate_constant import *

from idaes.core import MaterialFlowBasis
from idaes.core.util.testing import PhysicalParameterTestBlock


@pytest.fixture
def model():
    m = ConcreteModel()

    # # Add a test thermo package for validation
    m.pparams = PhysicalParameterTestBlock()
    m.thermo = m.pparams.build_state_block([1])

    m.rparams = GenericReactionParameterBlock(
        property_package=m.pparams,
        reaction_basis=MaterialFlowBasis.molar,
        rate_reactions={
            "r1": {
                "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                "rate_form": "foo",
                "concentration_form": ConcentrationForm.moleFraction,
            }
        },
        base_units={
            "amount": pyunits.mol,
            "mass": pyunits.kg,
            "time": pyunits.s,
            "length": pyunits.m,
            "temperature": pyunits.K,
        },
    )

    m.rxn = m.rparams.build_reaction_block(
        [1], state_block=m.thermo, has_equilibrium=False
    )

    return m


@pytest.mark.unit
def test_arrhenius_mole_frac(model):
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": 1,
        "energy_activation": 500,
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, pyunits.mol / pyunits.m**3 / pyunits.s)


@pytest.mark.unit
def test_arrhenius_mole_frac_convert(model):
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": (1e-3, pyunits.kmol / pyunits.m**3 / pyunits.s),
        "energy_activation": (0.5, pyunits.kJ / pyunits.mol),
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, pyunits.mol / pyunits.m**3 / pyunits.s)


@pytest.mark.unit
def test_arrhenius_molarity(model):
    model.rparams.config.rate_reactions.r1.concentration_form = (
        ConcentrationForm.molarity
    )
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": 1,
        "energy_activation": 500,
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, 1 / pyunits.s)


@pytest.mark.unit
def test_arrhenius_molarity_convert(model):
    model.rparams.config.rate_reactions.r1.concentration_form = (
        ConcentrationForm.molarity
    )
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": (3600, 1 / pyunits.hr),
        "energy_activation": (0.5, pyunits.kJ / pyunits.mol),
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, 1 / pyunits.s)


@pytest.mark.unit
def test_arrhenius_molality(model):
    model.rparams.config.rate_reactions.r1.concentration_form = (
        ConcentrationForm.molality
    )
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": 1,
        "energy_activation": 500,
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, pyunits.kg / pyunits.m**3 / pyunits.s)


@pytest.mark.unit
def test_arrhenius_molality_convert(model):
    model.rparams.config.rate_reactions.r1.concentration_form = (
        ConcentrationForm.molality
    )
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": (1e3, pyunits.g / pyunits.m**3 / pyunits.s),
        "energy_activation": (0.5, pyunits.kJ / pyunits.mol),
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, pyunits.kg / pyunits.m**3 / pyunits.s)


@pytest.mark.unit
def test_arrhenius_partial_pressure(model):
    model.rparams.config.rate_reactions.r1.concentration_form = (
        ConcentrationForm.partialPressure
    )
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": 1,
        "energy_activation": 500,
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, pyunits.mol / pyunits.m**3 / pyunits.s / pyunits.Pa)


@pytest.mark.unit
def test_arrhenius_partial_pressure_convert(model):
    model.rparams.config.rate_reactions.r1.concentration_form = (
        ConcentrationForm.partialPressure
    )
    model.rparams.config.rate_reactions.r1.parameter_data = {
        "arrhenius_const": (
            1e-3,
            pyunits.kmol / pyunits.m**3 / pyunits.s / pyunits.Pa,
        ),
        "energy_activation": (0.5, pyunits.kJ / pyunits.mol),
    }

    arrhenius.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.arrhenius_const, Var)
    assert model.rparams.reaction_r1.arrhenius_const.value == 1

    assert isinstance(model.rparams.reaction_r1.energy_activation, Var)
    assert model.rparams.reaction_r1.energy_activation.value == 500

    # Check expressions
    rform = arrhenius.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300 * pyunits.K
    )

    assert value(rform) == pytest.approx(0.81836, rel=1e-3)
    assert_units_equivalent(rform, pyunits.mol / pyunits.m**3 / pyunits.s / pyunits.Pa)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Make sure the supercritical steam cycle example solves.
"""

__author__ = "John Eslick"

import os
import pytest
import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.models_extra.power_generation.flowsheets.supercritical_steam_cycle.supercritical_steam_cycle import (
    main,
    pfd_result,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    activated_equalities_generator,
)
from idaes.models.properties import iapws95
from idaes.core.util.tables import create_stream_table_dataframe  # as Pandas DataFrame
from idaes.models.properties.general_helmholtz import helmholtz_available


@pytest.fixture(scope="module")
def model():
    m, solver = main()
    m.solver = solver

    return m


def gross_power_mw(model):
    # pyo.value(m.fs.turb.power[0]) is the power consumed in Watts
    return -pyo.value(model.fs.turb.power[0]) / 1e6


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_init(model):
    # check that the model solved properly and has 0 degrees of freedom
    assert degrees_of_freedom(model) == 0
    for c in activated_equalities_generator(model):
        assert abs(c.body() - c.lower) < 5e-4


# @pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
# @pytest.mark.integration
# def test_unit_consistency(model):
#     assert_units_consistent(model)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_init_value(model):
    assert gross_power_mw(model) == pytest.approx(622.38, abs=1e-2)
    # Make sure the stream table can be generated.
    df = create_stream_table_dataframe(streams=model._streams, orient="index")
    pfd_result(model, df, None)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_valve_change(model):
    model.fs.turb.throttle_valve[1].valve_opening[:].value = 0.25
    model.solver.solve(model, tee=True)
    assert gross_power_mw(model) == pytest.approx(594.66, abs=1e-2)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# =============================================================================
"""
SCPC Boiler subflowsheet

Inputs:
    BFW - boiler feed water (from Feed water heaters)
    Coal from pulverizers

Main Assumptions:
    Coal flowrate as a function of load, coal HHV is fixed and heat dutty
    split from fire side to water wall and platen superheater is fixed.

    Boiler heat exchanger network:
        Water Flow:
            BFW -> ECONOMIZER -> Water Wall -> Primary SH -> Platen SH -> Finishing Superheate -> HP Turbine -> Reheater -> IP Turbine
        Flue Gas Flow:
            Fire Ball -> Platen SH -> Finishing SH -> Reheater  -> o -> Economizer -> Air Preheater
                                                   -> Primary SH --^

        * HP Turbine, IP Turbine, Air Preheater ==> not included in this release

    Models used:
        - Mixers: Attemperator, Flue gas mix
        - Heater: Platen SH, Fire/Water side (simplified model)
        - BoilerHeatExchanger: Economizer, Primary SH, Finishing SH, Reheater
            + Shell and tube heat exchanger
                - tube side: Steam (side 1 holdup)
                - shell side: flue gas (side 2 holdup)

    Property packages used:
        - IAPWS: Water/steam side
        - IDEAL GAS: Flue Gas side

Created: 1/10/2020 by Boiler subsystem team (M Zamarripa)

"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

__author__ = "Miguel Zamarripa"

from collections import OrderedDict
import os
import logging

# Import Pyomo libraries
from pyomo.environ import (
    ConcreteModel,
    value,
    TransformationFactory,
    units as pyunits,
)
from pyomo.network import Arc
from pyomo.common.fileutils import this_file_dir

from idaes.core.util.tags import svg_tag

# Import IDAES core
from idaes.core import FlowsheetBlock

# Import Unit Model Modules
from idaes.models.properties import iapws95

# Import Property Modules
from idaes.models_extra.power_generation.properties import FlueGasParameterBlock

# Import Unit Model Modules
from idaes.models.unit_models import Heater, Mixer
from idaes.models_extra.power_generation.unit_models.boiler_heat_exchanger import (
    BoilerHeatExchanger,
    TubeArrangement,
    HeatExchangerFlowPattern,
)
from idaes.models.unit_models.separator import (
    Separator,
    SplittingType,
    EnergySplittingType,
)

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.prop_water = iapws95.Iapws95ParameterBlock()

    build_boiler(m.fs)

    # Create a solver
    solver = get_solver()
    return (m, solver)


def build_boiler(fs):
    # Add property packages to flowsheet library
    fs.prop_fluegas = FlueGasParameterBlock()

    # Create unit models
    # Boiler Economizer
    fs.ECON = BoilerHeatExchanger(
        cold_side={"property_package": fs.prop_water, "has_pressure_change": True},
        hot_side={"property_package": fs.prop_fluegas, "has_pressure_change": True},
        has_holdup=False,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        tube_arrangement=TubeArrangement.inLine,
        cold_side_water_phase="Liq",
        has_radiation=False,
    )
    # Primary Superheater
    fs.PrSH = BoilerHeatExchanger(
        cold_side={"property_package": fs.prop_water, "has_pressure_change": True},
        hot_side={"property_package": fs.prop_fluegas, "has_pressure_change": True},
        has_holdup=False,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        tube_arrangement=TubeArrangement.inLine,
        cold_side_water_phase="Vap",
        has_radiation=True,
    )

    # Finishing Superheater
    fs.FSH = BoilerHeatExchanger(
        cold_side={"property_package": fs.prop_water, "has_pressure_change": True},
        hot_side={"property_package": fs.prop_fluegas, "has_pressure_change": True},
        has_holdup=False,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        tube_arrangement=TubeArrangement.inLine,
        cold_side_water_phase="Vap",
        has_radiation=True,
    )

    # Reheater
    fs.RH = BoilerHeatExchanger(
        cold_side={"property_package": fs.prop_water, "has_pressure_change": True},
        hot_side={"property_package": fs.prop_fluegas, "has_pressure_change": True},
        has_holdup=False,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        tube_arrangement=TubeArrangement.inLine,
        cold_side_water_phase="Vap",
        has_radiation=True,
    )
    # Platen Superheater
    fs.PlSH = Heater(property_package=fs.prop_water)

    # Boiler Water Wall
    fs.Water_wall = Heater(property_package=fs.prop_water)

    # Boiler Splitter (splits FSH flue gas outlet to Reheater and PrSH)
    fs.Spl1 = Separator(
        property_package=fs.prop_fluegas,
        split_basis=SplittingType.totalFlow,
        energy_split_basis=EnergySplittingType.equal_temperature,
    )
    # Flue gas mixer (mixing FG from Reheater and Primary SH, inlet to ECON)
    fs.mix1 = Mixer(
        property_package=fs.prop_fluegas,
        inlet_list=["Reheat_out", "PrSH_out"],
        dynamic=False,
    )

    # Mixer for Attemperator #1 (between PrSH and PlSH)
    fs.ATMP1 = Mixer(
        property_package=fs.prop_water,
        inlet_list=["Steam", "SprayWater"],
        dynamic=False,
    )

    # Build connections (streams)

    # Steam Route (side 1 = tube side = steam/water side)
    # Boiler feed water to Economizer (to be imported in full plant model)
    #    fs.bfw2econ = Arc(source=fs.FWH8.outlet,
    #                           destination=fs.ECON.cold_side_inlet)
    fs.econ2ww = Arc(source=fs.ECON.cold_side_outlet, destination=fs.Water_wall.inlet)
    fs.ww2prsh = Arc(source=fs.Water_wall.outlet, destination=fs.PrSH.cold_side_inlet)
    fs.prsh2plsh = Arc(source=fs.PrSH.cold_side_outlet, destination=fs.PlSH.inlet)
    fs.plsh2fsh = Arc(source=fs.PlSH.outlet, destination=fs.FSH.cold_side_inlet)
    fs.FSHtoATMP1 = Arc(source=fs.FSH.cold_side_outlet, destination=fs.ATMP1.Steam)
    #    fs.fsh2hpturbine=Arc(source=fs.ATMP1.outlet,
    #                           destination=fs.HPTinlet)
    # (to be imported in full plant model)

    # Flue gas route ---------------------------------------------------------
    # water wall connected with boiler block (to fix the heat duty)
    # platen SH connected with boiler block (to fix the heat duty)
    # Finishing superheater connected with a flowsheet level constraint
    fs.fg_fsh2_separator = Arc(source=fs.FSH.hot_side_outlet, destination=fs.Spl1.inlet)
    fs.fg_fsh2rh = Arc(source=fs.Spl1.outlet_1, destination=fs.RH.hot_side_inlet)
    fs.fg_fsh2PrSH = Arc(source=fs.Spl1.outlet_2, destination=fs.PrSH.hot_side_inlet)
    fs.fg_rhtomix = Arc(source=fs.RH.hot_side_outlet, destination=fs.mix1.Reheat_out)
    fs.fg_prsh2mix = Arc(source=fs.PrSH.hot_side_outlet, destination=fs.mix1.PrSH_out)
    fs.fg_mix2econ = Arc(source=fs.mix1.outlet, destination=fs.ECON.hot_side_inlet)

    TransformationFactory("network.expand_arcs").apply_to(fs)


# Set inputs ==========================
def initialize(m):
    # ------------- ECONOMIZER -----------------------------------------------
    # BFW Boiler Feed Water inlet temeperature = 555 F = 563.706 K
    m.fs.ECON.cold_side_inlet.flow_mol[0].fix(24194.177)  # mol/s
    m.fs.ECON.cold_side_inlet.enth_mol[0].fix(14990.97)
    m.fs.ECON.cold_side_inlet.pressure[0].fix(26922222.222)  # Pa

    # FLUE GAS Inlet from Primary Superheater
    FGrate = 21290.6999  # mol/s
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.ECON.hot_side_inlet.temperature[0].fix(682.335)  # K
    m.fs.ECON.hot_side_inlet.pressure[0].fix(100145)  # Pa

    # economizer design variables and parameters
    ITM = 0.0254  # inch to meter conversion
    # Based on NETL Baseline Report Rev3
    m.fs.ECON.tube_di.fix((2 - 2 * 0.188) * ITM)  # calc inner diameter
    #                                   (2 = outer diameter, thickness = 0.188)
    m.fs.ECON.tube_thickness.fix(0.188 * ITM)  # tube thickness
    m.fs.ECON.pitch_x.fix(3.5 * ITM)
    # pitch_y = (54.5) gas path transverse width /columns
    m.fs.ECON.pitch_y.fix(5.03 * ITM)
    m.fs.ECON.tube_length.fix(53.41 * 12 * ITM)  # use tube length (53.41 ft)
    m.fs.ECON.tube_nrow.fix(36 * 2.5)  # use to match baseline performance
    m.fs.ECON.tube_ncol.fix(130)  # 130 from NETL
    m.fs.ECON.nrow_inlet.fix(2)
    m.fs.ECON.delta_elevation.fix(50)
    # parameters
    # heat transfer resistance due to tube side fouling (water scales)
    m.fs.ECON.tube_r_fouling = 0.000176
    # heat transfer resistance due to tube shell fouling (ash deposition)
    m.fs.ECON.shell_r_fouling = 0.00088
    if m.fs.ECON.config.has_radiation is True:
        m.fs.ECON.emissivity_wall.fix(0.7)  # wall emissivity
    # correction factor for overall heat transfer coefficient
    m.fs.ECON.fcorrection_htc.fix(1.5)
    # correction factor for pressure drop calc tube side
    m.fs.ECON.fcorrection_dp_tube.fix(1.0)
    # correction factor for pressure drop calc shell side
    m.fs.ECON.fcorrection_dp_shell.fix(1.0)

    # --------- Primary Superheater ------------
    # Steam from water wall
    hprsh = value(iapws95.htpx(773.15 * pyunits.K, 24865516.722 * pyunits.Pa))
    print(hprsh)
    m.fs.PrSH.cold_side_inlet.flow_mol[0].fix(24190.26)  # mol/s
    m.fs.PrSH.cold_side_inlet.enth_mol[0].fix(hprsh)  # J/mol
    m.fs.PrSH.cold_side_inlet.pressure[0].fix(2.5449e7)  # Pascals

    # FLUE GAS Inlet from Primary Superheater
    FGrate = 21290.6999 * 0.18  # mol/s
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.PrSH.hot_side_inlet.temperature[0].fix(1180.335)
    m.fs.PrSH.hot_side_inlet.pressure[0].fix(100145)

    # Primary Superheater
    ITM = 0.0254  # inch to meter conversion
    m.fs.PrSH.tube_di.fix((2.5 - 2 * 0.165) * ITM)
    m.fs.PrSH.tube_thickness.fix(0.165 * ITM)
    m.fs.PrSH.pitch_x.fix(3 * ITM)
    # gas path transverse width 54.78 ft / number of columns
    m.fs.PrSH.pitch_y.fix(54.78 / 108 * 12 * ITM)
    m.fs.PrSH.tube_length.fix(53.13 * 12 * ITM)
    m.fs.PrSH.tube_nrow.fix(20 * 2)
    m.fs.PrSH.tube_ncol.fix(108)
    m.fs.PrSH.nrow_inlet.fix(4)
    m.fs.PrSH.delta_elevation.fix(50)
    m.fs.PrSH.tube_r_fouling = 0.000176  # (0.001 h-ft^2-F/BTU)
    m.fs.PrSH.shell_r_fouling = 0.003131  # (0.03131 - 0.1779 h-ft^2-F/BTU)
    if m.fs.PrSH.config.has_radiation is True:
        m.fs.PrSH.emissivity_wall.fix(0.7)  # wall emissivity
    # correction factor for overall heat transfer coefficient
    m.fs.PrSH.fcorrection_htc.fix(1.5)
    # correction factor for pressure drop calc tube side
    m.fs.PrSH.fcorrection_dp_tube.fix(1.0)
    # correction factor for pressure drop calc shell side
    m.fs.PrSH.fcorrection_dp_shell.fix(1.0)

    #  ----- Finishing Superheater ----------------------
    # Steam from Platen Supeheater
    hfsh = value(iapws95.htpx(823.15 * pyunits.K, 24790249.01 * pyunits.Pa))
    m.fs.FSH.cold_side_inlet.flow_mol[0].fix(24194.177)  # mol/s
    m.fs.FSH.cold_side_inlet.enth_mol[0].fix(hfsh)  # J/mol
    m.fs.FSH.cold_side_inlet.pressure[0].fix(24790249.01)  # Pascals

    # FLUE GAS Inlet from Primary Superheater
    FGrate = 21290.6999  # mol/s
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.FSH.hot_side_inlet.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.FSH.hot_side_inlet.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.FSH.hot_side_inlet.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.FSH.hot_side_inlet.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.FSH.hot_side_inlet.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.FSH.hot_side_inlet.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.FSH.hot_side_inlet.temperature[0].fix(1300.335)
    m.fs.FSH.hot_side_inlet.pressure[0].fix(100145)

    # Finishing Superheater
    ITM = 0.0254  # inch to meter conversion
    m.fs.FSH.tube_di.fix((2.5 - 2 * 0.165) * ITM)
    m.fs.FSH.tube_thickness.fix(0.165 * ITM)
    m.fs.FSH.pitch_x.fix(3 * ITM)
    # gas path transverse width 54.78 ft / number of columns
    m.fs.FSH.pitch_y.fix(54.78 / 108 * 12 * ITM)
    m.fs.FSH.tube_length.fix(53.13 * 12 * ITM)
    m.fs.FSH.tube_nrow.fix(8 * 2)
    m.fs.FSH.tube_ncol.fix(85)
    m.fs.FSH.nrow_inlet.fix(2)
    m.fs.FSH.delta_elevation.fix(50)
    m.fs.FSH.tube_r_fouling = 0.000176  # (0.001 h-ft^2-F/BTU)
    m.fs.FSH.shell_r_fouling = 0.003131  # (0.03131 - 0.1779 h-ft^2-F/BTU)
    if m.fs.FSH.config.has_radiation is True:
        m.fs.FSH.emissivity_wall.fix(0.7)  # wall emissivity
    # correction factor for overall heat transfer coefficient
    m.fs.FSH.fcorrection_htc.fix(1.0)
    # correction factor for pressure drop calc tube side
    m.fs.FSH.fcorrection_dp_tube.fix(1.0)
    # correction factor for pressure drop calc shell side
    m.fs.FSH.fcorrection_dp_shell.fix(1.0)

    #  ----- Reheater Superheater ----------------------
    #   Steam from HP Turbine outlet
    m.fs.RH.cold_side_inlet.flow_mol[0].fix(21235.27)  # mol/s
    m.fs.RH.cold_side_inlet.enth_mol[0].fix(53942.7569)  # J/mol
    m.fs.RH.cold_side_inlet.pressure[0].fix(3677172.33638)  # Pascals

    # FLUE GAS Inlet from Finishing Superheater
    FGrate = 21290.6999 * 0.85  # mol/s
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.RH.hot_side_inlet.temperature[0].fix(1180.335)
    m.fs.RH.hot_side_inlet.pressure[0].fix(100145)

    # Reheater Superheater
    ITM = 0.0254  # inch to meter conversion
    m.fs.RH.tube_di.fix((2.5 - 2 * 0.165) * ITM)
    m.fs.RH.tube_thickness.fix(0.11 * ITM)
    m.fs.RH.pitch_x.fix(3 * ITM)
    # gas path transverse width 54.08 ft / number of columns
    m.fs.RH.pitch_y.fix(54.08 / 108 * 12 * ITM)
    m.fs.RH.tube_length.fix(53.82 * 12 * ITM)
    m.fs.RH.tube_nrow.fix(18 * 2)
    m.fs.RH.tube_ncol.fix(82 + 70)
    m.fs.RH.nrow_inlet.fix(2)
    m.fs.RH.delta_elevation.fix(30)
    m.fs.RH.tube_r_fouling = 0.000176  # (0.001 h-ft^2-F/BTU)
    m.fs.RH.shell_r_fouling = 0.00088  # (0.03131 - 0.1779 h-ft^2-F/BTU)
    if m.fs.RH.config.has_radiation is True:
        m.fs.RH.emissivity_wall.fix(0.7)  # wall emissivity
    # correction factor for overall heat transfer coefficient
    m.fs.RH.fcorrection_htc.fix(1.8)
    # correction factor for pressure drop calc tube side
    m.fs.RH.fcorrection_dp_tube.fix(1.0)
    # correction factor for pressure drop calc shell side
    m.fs.RH.fcorrection_dp_shell.fix(1.0)

    #    Platen Superheater ------------------------------------------
    hpl = value(iapws95.htpx(798.15 * pyunits.K, 24790249.01 * pyunits.Pa))
    m.fs.PlSH.inlet[:].flow_mol.fix(24194.177)
    m.fs.PlSH.inlet[:].enth_mol.fix(hpl)
    m.fs.PlSH.inlet[:].pressure.fix(24790249.01)
    m.fs.PlSH.heat_duty[:].fix(5.5e7)

    #    Water wall Superheater ------------------------------------------
    hww = value(iapws95.htpx(588.15 * pyunits.K, 2.5449e7 * pyunits.Pa))
    m.fs.Water_wall.inlet[:].flow_mol.fix(24194.177)
    m.fs.Water_wall.inlet[:].enth_mol.fix(hww)
    m.fs.Water_wall.inlet[:].pressure.fix(24865516.722)
    m.fs.Water_wall.heat_duty[:].fix(7.51e8)  # 8.76e8

    #   splitter flue gas from Finishing SH to Reheater and Primary SH
    m.fs.Spl1.split_fraction[0, "outlet_1"].fix(0.75)  # 0.85)
    # FLUE GAS Inlet from Primary Superheater
    FGrate = 21290.6999  # mol/s
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.Spl1.inlet.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.Spl1.inlet.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.Spl1.inlet.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.Spl1.inlet.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.Spl1.inlet.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.Spl1.inlet.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.Spl1.inlet.temperature[0].fix(1180.335)
    m.fs.Spl1.inlet.pressure[0].fix(100145)

    # mixer (econ inlet) fluegas outlet from reheater and primary superheater
    #   Mixer inlets  ['Reheat_out', 'PrSH_out']
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.mix1.Reheat_out.pressure.fix(100145)
    m.fs.mix1.Reheat_out.temperature.fix(731.5)
    # Fixed SprayWater as small flow and arbitrary conditions
    # (will be connected from FW pump splitter)
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.mix1.PrSH_out.pressure.fix(100145)
    m.fs.mix1.PrSH_out.temperature.fix(731.15)

    # Attemperator inputs
    hatt = value(iapws95.htpx(875.15 * pyunits.K, 24865516.722 * pyunits.Pa))
    m.fs.ATMP1.Steam.flow_mol.fix(24194.177)
    m.fs.ATMP1.Steam.enth_mol.fix(hatt)
    m.fs.ATMP1.Steam.pressure.fix(24865516.722)
    # Fixed SprayWater from FW pump splitter (splitter is needded)
    hatt2 = value(iapws95.htpx(563.15 * pyunits.K, 2.5449e7 * pyunits.Pa))
    m.fs.ATMP1.SprayWater_state[:].flow_mol.fix(0.001)
    m.fs.ATMP1.SprayWater_state[:].pressure.fix(1.22e8)
    m.fs.ATMP1.SprayWater_state[:].enth_mol.fix(hatt2)

    # ---------  Initialization ------------------------------------------

    # Initialize Units
    m.fs.ECON.initialize(outlvl=logging.INFO)
    m.fs.PrSH.initialize(outlvl=logging.INFO)
    m.fs.FSH.initialize(outlvl=logging.INFO)
    m.fs.RH.initialize(outlvl=logging.INFO)
    m.fs.PlSH.initialize(outlvl=logging.INFO)
    m.fs.Water_wall.initialize(outlvl=logging.INFO)
    m.fs.Spl1.initialize(outlvl=logging.INFO)
    m.fs.mix1.initialize(outlvl=logging.INFO)
    m.fs.ATMP1.initialize(outlvl=logging.INFO)
    print("initialization done")


def unfix_inlets(m):
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "H2O"].unfix()
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "CO2"].unfix()
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "N2"].unfix()
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "O2"].unfix()
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "NO"].unfix()
    m.fs.ECON.hot_side_inlet.flow_mol_comp[0, "SO2"].unfix()
    m.fs.ECON.hot_side_inlet.temperature[0].unfix()
    m.fs.ECON.hot_side_inlet.pressure[0].unfix()

    # PrSH Primary superheater inlets (steam and flue gas) --------------------
    m.fs.PrSH.cold_side_inlet.flow_mol.unfix()
    m.fs.PrSH.cold_side_inlet.enth_mol[0].unfix()
    m.fs.PrSH.cold_side_inlet.pressure[0].unfix()

    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "H2O"].unfix()
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "CO2"].unfix()
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "N2"].unfix()
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "O2"].unfix()
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "NO"].unfix()
    m.fs.PrSH.hot_side_inlet.flow_mol_comp[0, "SO2"].unfix()
    m.fs.PrSH.hot_side_inlet.temperature[0].unfix()
    m.fs.PrSH.hot_side_inlet.pressure[0].unfix()

    # WaterWall  water from economizer  ---------------------------------------
    m.fs.Water_wall.inlet[:].flow_mol.unfix()
    m.fs.Water_wall.inlet[:].enth_mol.unfix()
    m.fs.Water_wall.inlet[:].pressure.unfix()

    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "H2O"].unfix()
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "CO2"].unfix()
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "N2"].unfix()
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "O2"].unfix()
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "NO"].unfix()
    m.fs.RH.hot_side_inlet.flow_mol_comp[0, "SO2"].unfix()
    m.fs.RH.hot_side_inlet.temperature[0].unfix()
    m.fs.RH.hot_side_inlet.pressure[0].unfix()

    # Finishing Superheater (steam from Platen SH)-----------------------------
    m.fs.FSH.cold_side_inlet.flow_mol.unfix()
    m.fs.FSH.cold_side_inlet.enth_mol[0].unfix()
    m.fs.FSH.cold_side_inlet.pressure[0].unfix()

    # Platen SH steam inlet conditions from Water Wall-------------------------
    m.fs.PlSH.inlet[:].flow_mol.unfix()
    m.fs.PlSH.inlet[:].enth_mol.unfix()
    m.fs.PlSH.inlet[:].pressure.unfix()

    # unfix Splitter flue gas inlet to RH and PrSH
    m.fs.Spl1.inlet.flow_mol_comp[0, "H2O"].unfix()
    m.fs.Spl1.inlet.flow_mol_comp[0, "CO2"].unfix()
    m.fs.Spl1.inlet.flow_mol_comp[0, "N2"].unfix()
    m.fs.Spl1.inlet.flow_mol_comp[0, "O2"].unfix()
    m.fs.Spl1.inlet.flow_mol_comp[0, "NO"].unfix()
    m.fs.Spl1.inlet.flow_mol_comp[0, "SO2"].unfix()
    m.fs.Spl1.inlet.temperature[0].unfix()
    m.fs.Spl1.inlet.pressure[0].unfix()

    # unfix mixer inlets (now connected)
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "H2O"].unfix()
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "CO2"].unfix()
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "N2"].unfix()
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "O2"].unfix()
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "NO"].unfix()
    m.fs.mix1.Reheat_out.flow_mol_comp[0, "SO2"].unfix()
    m.fs.mix1.Reheat_out.pressure.unfix()
    m.fs.mix1.Reheat_out.temperature.unfix()
    # PrSH output
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "H2O"].unfix()
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "CO2"].unfix()
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "N2"].unfix()
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "O2"].unfix()
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "NO"].unfix()
    m.fs.mix1.PrSH_out.flow_mol_comp[0, "SO2"].unfix()
    m.fs.mix1.PrSH_out.pressure.unfix()
    m.fs.mix1.PrSH_out.temperature.unfix()

    # unfix attemperator inlet from (Finishing SH)
    m.fs.ATMP1.Steam.flow_mol.unfix()
    m.fs.ATMP1.Steam.enth_mol.unfix()
    m.fs.ATMP1.Steam.pressure.unfix()


def pfd_result(outfile, m, df):
    tags = {}
    for i in df.index:
        tags[i + "_F"] = df.loc[i, "Molar Flow"]
        tags[i + "_T"] = df.loc[i, "T"]
        tags[i + "_P"] = df.loc[i, "P"]
        tags[i + "_X"] = df.loc[i, "Vapor Fraction"]

    tags["FG_2_RH_Fm"] = value(m.fs.RH.side_2.properties_in[0].flow_mass)
    tags["FG_2_RH_T"] = value(m.fs.RH.side_2.properties_in[0].temperature)
    tags["FG_2_RH_P"] = value(m.fs.RH.side_2.properties_in[0].pressure)

    tags["FG_RH_2_Mix_Fm"] = value(m.fs.RH.side_2.properties_out[0].flow_mass)
    tags["FG_RH_2_Mix_T"] = value(m.fs.RH.side_2.properties_out[0].temperature)
    tags["FG_RH_2_Mix_P"] = value(m.fs.RH.side_2.properties_out[0].pressure)

    tags["FG_2_FSH_Fm"] = value(m.fs.FSH.side_2.properties_in[0].flow_mass)
    tags["FG_2_FSH_T"] = value(m.fs.FSH.side_2.properties_in[0].temperature)
    tags["FG_2_FSH_P"] = value(m.fs.FSH.side_2.properties_in[0].pressure)

    tags["FG_2_PrSH_Fm"] = value(m.fs.PrSH.side_2.properties_in[0].flow_mass)
    tags["FG_2_PrSH_T"] = value(m.fs.PrSH.side_2.properties_in[0].temperature)
    tags["FG_2_PrSH_P"] = value(m.fs.PrSH.side_2.properties_in[0].pressure)

    tags["FG_PrSH_2_Mix_Fm"] = value(m.fs.PrSH.side_2.properties_out[0].flow_mass)
    tags["FG_PrSH_2_Mix_T"] = value(m.fs.PrSH.side_2.properties_out[0].temperature)
    tags["FG_PrSH_2_Mix_P"] = value(m.fs.PrSH.side_2.properties_out[0].pressure)

    tags["FG_2_ECON_Fm"] = value(m.fs.ECON.side_2.properties_in[0].flow_mass)
    tags["FG_2_ECON_T"] = value(m.fs.ECON.side_2.properties_in[0].temperature)
    tags["FG_2_ECON_P"] = value(m.fs.ECON.side_2.properties_in[0].pressure)

    tags["FG_2_AIRPH_Fm"] = value(m.fs.ECON.side_2.properties_out[0].flow_mass)
    tags["FG_2_AIRPH_T"] = value(m.fs.ECON.side_2.properties_out[0].temperature)
    tags["FG_2_AIRPH_P"] = value(m.fs.ECON.side_2.properties_out[0].pressure)

    tags["FG_2_STACK_Fm"] = value(m.fs.ECON.side_2.properties_out[0].flow_mass)
    tags["FG_2_STACK_T"] = value(m.fs.ECON.side_2.properties_out[0].temperature)
    tags["FG_2_STACK_P"] = value(m.fs.ECON.side_2.properties_out[0].pressure)

    original_svg_file = os.path.join(this_file_dir(), "Boiler_scpc_PFD.svg")
    with open(original_svg_file, "r") as f:
        svg_tag(tags, f, outfile=outfile)


def _stream_dict(m):
    """Adds _streams to m, which contains a dictionary of streams for display

    Args:
        m (ConcreteModel): A Pyomo model from create_model()

    Returns:
        None
    """
    # We control m
    # pylint: disable-next=protected-access
    m._streams = OrderedDict(
        [
            ("MS", m.fs.ATMP1.mixed_state),
            ("ATMP_In", m.fs.FSH.side_1.properties_out),
            ("FSH_In", m.fs.FSH.side_1.properties_in),
            ("PrSH_IN", m.fs.PrSH.side_1.properties_in),
            ("RHT_COLD", m.fs.RH.side_1.properties_in),
            ("RHT_HOT", m.fs.RH.side_1.properties_out),
            ("PlatenSH_IN", m.fs.PlSH.control_volume.properties_in),
            ("BFW", m.fs.ECON.side_1.properties_in),
            ("ECON_OUT", m.fs.ECON.side_1.properties_out),
        ]
    )


def print_results(m):
    print()
    print("Results")
    print()

    print("viscosity gas side = ", m.fs.PrSH.side_2.properties_in[0].visc_d.value)
    print(
        "conductivity gas side = ", m.fs.PrSH.side_2.properties_in[0].therm_cond.value
    )
    print("velocity_tube = ", m.fs.PrSH.v_tube[0].value)
    print("velocity_shell = ", m.fs.PrSH.v_shell[0].value)
    print("Re_tube = ", m.fs.PrSH.N_Re_tube[0].value)
    print("Re_shell = ", m.fs.PrSH.N_Re_shell[0].value)
    print("hconv_tube = ", m.fs.PrSH.hconv_tube[0].value)
    print("hconv_shell_rad = ", m.fs.PrSH.hconv_shell_rad[0].value)
    print("hconv_shell_conv = ", m.fs.PrSH.hconv_shell_conv[0].value)
    print("hconv_shell_total = ", m.fs.PrSH.hconv_shell_total[0].value)
    print("driving force = ", m.fs.PrSH.temperature_driving_force[0].value)
    print("dT_inlet = ", m.fs.PrSH.deltaT_1[0].value)
    print("dT_outlet = ", m.fs.PrSH.deltaT_2[0].value)
    print("deltaP tube = ", m.fs.PrSH.deltaP_tube[0].value)
    print("deltaP shell = ", m.fs.PrSH.deltaP_shell[0].value)
    print("mbl = ", value(m.fs.PrSH.mbl))

    if m.fs.PrSH.config.has_radiation is True:
        print("gas emissivity = ", m.fs.PrSH.gas_emissivity[0].value)
        print("gas emissivity div2 = ", m.fs.PrSH.gas_emissivity_div2[0].value)
        print("gas emissivity mul2 = ", m.fs.PrSH.gas_emissivity_mul2[0].value)
        print("gas gray fraction = ", m.fs.PrSH.gas_gray_fraction[0].value)
    print(
        "liquid density in = ",
        value(m.fs.PrSH.side_1.properties_in[0].dens_mass_phase["Liq"]),
    )

    print(
        "liquid density out = ",
        value(m.fs.PrSH.side_1.properties_out[0].dens_mass_phase["Liq"]),
    )
    print("heat transfer area = ", value(m.fs.PrSH.area_heat_transfer))
    print(
        "overall heat transfer = ",
        value(m.fs.PrSH.overall_heat_transfer_coefficient[0]),
    )

    print("\n\n ------------- Economizer   ---------")
    print("liquid temp in = ", value(m.fs.ECON.side_1.properties_in[0].temperature))
    print("liquid temp out = ", value(m.fs.ECON.side_1.properties_out[0].temperature))
    print("gas temp in = ", value(m.fs.ECON.side_2.properties_in[0].temperature))
    print("gas temp out = ", value(m.fs.ECON.side_2.properties_out[0].temperature))

    print("\n\n ------------- water wall  ---------")
    print(
        "liquid temp in = ",
        value(m.fs.Water_wall.control_volume.properties_in[0].temperature),
    )
    print(
        "steam temp out = ",
        value(m.fs.Water_wall.control_volume.properties_out[0].temperature),
    )

    print("\n\n ------------- Primary Superheater  ---------")
    print("steam temp in = ", value(m.fs.PrSH.side_1.properties_in[0].temperature))
    print("steam temp out = ", value(m.fs.PrSH.side_1.properties_out[0].temperature))
    print("gas temp in = ", value(m.fs.PrSH.side_2.properties_in[0].temperature))
    print("gas temp out = ", value(m.fs.PrSH.side_2.properties_out[0].temperature))

    print("\n\n ------------- Platen SH  ---------")
    print(
        "steam temp in = ", value(m.fs.PlSH.control_volume.properties_in[0].temperature)
    )
    print(
        "steam temp out = ",
        value(m.fs.PlSH.control_volume.properties_out[0].temperature),
    )

    print("\n\n ------------- Finishing Superheater  ---------")
    print("steam temp in = ", value(m.fs.FSH.side_1.properties_in[0].temperature))
    print(
        "steam temp out (to attmp) = ",
        value(m.fs.FSH.side_1.properties_out[0].temperature),
    )
    print("gas temp in = ", value(m.fs.FSH.side_2.properties_in[0].temperature))
    print("gas temp out = ", value(m.fs.FSH.side_2.properties_out[0].temperature))

    print("\n\n ------------- Attemperator  ---------")
    print(
        "steam temp in = ",
        value(m.fs.ATMP1.Steam.enth_mol[0]),
        value(m.fs.FSH.side_1.properties_out[0].temperature),
    )
    print(
        "steam temp out (to HP turbine) = ",
        value(m.fs.ATMP1.outlet.enth_mol[0]),
        value(m.fs.ATMP1.mixed_state[0].temperature),
    )
    print()

    print("\n\n ------------- Reheater  ---------")
    print("liquid temp in = ", value(m.fs.RH.side_1.properties_in[0].temperature))
    print("liquid temp out = ", value(m.fs.RH.side_1.properties_out[0].temperature))
    print("gas temp in = ", value(m.fs.RH.side_2.properties_in[0].temperature))
    print("gas temp out = ", value(m.fs.RH.side_2.properties_out[0].temperature))


if __name__ == "__main__":
    m, solver = main()

    solver.options = {
        "tol": 1e-6,
        "linear_solver": "ma27",
        "max_iter": 100,
        "halt_on_ampl_error": "yes",
    }
    # initialize each unit at the time
    initialize(m)
    print("flowsheet degrees of freedom = " + str(degrees_of_freedom(m)))
    # unfix inlets to build arcs at the flowsheet level
    unfix_inlets(m)
    # users can change the following fixed values to obtain certain performance
    # m.fs.ATMP1.outlet.enth_mol[0].fix(62710.01)
    # m.fs.Water_wall.heat_duty.unfix()
    # m.fs.RH.side_1_outlet.enth_mol.fix(66043.35)
    # m.fs.PlSH.heat_duty.unfix()
    #    m.fs.ATMP1.SprayWater.flow_mol[0].unfix()
    print("flowsheet degrees of freedom = " + str(degrees_of_freedom(m)))
    results = solver.solve(m, tee=True, symbolic_solver_labels=True)
    print_results(m)

    #    print results in the PFD file (activate if needed)
    #    _stream_dict(m)
    #    df = create_stream_table_dataframe(streams=m._streams, orient="index")
    #    pfd_result("Boiler_Results.svg", m, df)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains utility functions to generate phase equilibrium data and
plots.
"""
# Import plotting functions
import matplotlib.pyplot as plt
import numpy as np

# Import objects from pyomo package
from pyomo.environ import (
    check_optimal_termination,
    value,
    units as pyunits,
)

from idaes.core.solvers import get_solver
import idaes.logger as idaeslog


__author__ = "Alejandro Garciadiego"


def Txy_diagram(
    model,
    component_1,
    component_2,
    pressure,
    num_points=20,
    temperature=298.15,
    figure_name=None,
    print_legend=True,
    include_pressure=False,
    print_level=idaeslog.NOTSET,
    solver=None,
    solver_op=None,
):
    """
    This method generates T-x-y plots. Given the components, pressure and property dictionary
    this function calls Txy_data() to generate T-x-y data and once the data has
    been generated calls build_txy_diagrams() to create a plot.

    Args:
        component_1: Component which composition will be plotted in x axis
        component_2: Component which composition will decrease in x axis
        pressure: Pressure at which the bubble and drew temperatures will be calculated
        temperature: Temperature at which to initialize state block
        num_points: Number of data point to be calculated
        properties: property package which contains parameters to calculate bubble
        and dew temperatures for the mixture of the components specified.
        figure_name: if a figure name is included the plot will save with the name
        figure_name.png
        print_legend (bool): = If True, include legend to distinguish between
            Bubble and dew temperature. The default is True.
        include_pressure (bool) = If True, print pressure at which the plot is
        calculated in legends. The default is False.
        print_level: printing level from initialization
        solver: solver to use (default=None, use IDAES default solver)
        solver_op: solver options

    Returns:
        Plot
    """
    # Run txy_ data function to obtain bubble and dew twmperatures
    Txy_data_to_plot = Txy_data(
        model,
        component_1,
        component_2,
        pressure,
        num_points,
        temperature,
        print_level,
        solver,
        solver_op,
    )

    # Run diagrams function to convert t-x-y data into a plot
    build_txy_diagrams(Txy_data_to_plot, figure_name, print_legend, include_pressure)


def Txy_data(
    model,
    component_1,
    component_2,
    pressure,
    num_points=20,
    temperature=298.15,
    print_level=idaeslog.NOTSET,
    solver="ipopt_v2",
    solver_op=None,
    solver_writer_config=None,
):
    """
    Function to generate T-x-y data. The function builds a state block and extracts
    bubble and dew temperatures at P pressure for N number of compositions.
    As N is increased the time of the calculation will increase and
    create a smoother looking plot.

    Args:
        component_1: Component 1
        component_2: Component 2
        pressure: Pressure at which the bubble and drew temperatures will be calculated
        temperature: Temperature at which to initialize state block
        num_points: Number of data point to be calculated
        model: Model with initialized property package which contains data to calculate
        bubble and dew temperatures for  component 1 and component 2
        print_level: printing level from initialization
        solver: solver to use (default=None, use IDAES default solver)
        solver_op: solver options
        solver_writer_config: writer_config arguments for solver

    Returns:
        (Class): A class containing the T-x-y data

    """

    components = list(model.params.component_list)
    components_used = [component_1, component_2]
    components_not_used = list(set(components) - set(components_used))

    # Add properties parameter blocks to the flowsheet with specifications

    model.props = model.params.build_state_block([1], defined_state=True)

    # Set initial concentration of component 1 close to 1
    x = 0.99

    # Set conditions for flash unit model
    model.props[1].mole_frac_comp[component_1].fix(x)

    for i in components_not_used:
        model.props[1].mole_frac_comp[i].fix(1e-5)

    xs = sum(value(model.props[1].mole_frac_comp[i]) for i in components_not_used)

    model.props[1].mole_frac_comp[component_2].fix(1 - x - xs)

    model.props[1].flow_mol.fix(1)
    model.props[1].temperature.fix(temperature)
    model.props[1].pressure.fix(pressure)

    # Initialize flash unit model
    model.props[1].calculate_scaling_factors()
    # TODO: This will need to be updated at some point as we deprecate the old initialization API
    model.props.initialize(solver=solver, optarg=solver_op, outlvl=print_level)

    solver = get_solver(
        solver, solver_options=solver_op, writer_config=solver_writer_config
    )

    # Create an array of compositions with N number of points
    x_d = np.linspace(x, 1 - x - xs, num_points)

    # Create empty arrays for concentration, bubble temperature and dew temperature
    X = []
    Tbubb = []
    Tdew = []

    # Obtain pressure and temperature units from the unit model
    Punit = pyunits.get_units(model.props[1].pressure)
    Tunit = pyunits.get_units(model.props[1].temperature)

    count = 1
    # Create and run loop to calculate temperatures at every composition
    for i, v in enumerate(x_d):
        model.props[1].mole_frac_comp[component_1].fix(v)
        model.props[1].mole_frac_comp[component_2].fix(1 - v - xs)
        # solve the model
        status = solver.solve(model, tee=False)
        # If solution is optimal store the concentration, and calculated temperatures in the created arrays
        if check_optimal_termination(status):

            print(f"Case: {count} Optimal. {component_1} x = {v:.2f}")

            if hasattr(model.props[1], "_mole_frac_tdew") and hasattr(
                model.props[1], "_mole_frac_tbub"
            ):
                Tbubb.append(value(model.props[1].temperature_bubble["Vap", "Liq"]))
                Tdew.append(value(model.props[1].temperature_dew["Vap", "Liq"]))

            elif hasattr(model.props[1], "_mole_frac_tdew"):
                print("One of the components only exists in vapor phase.")
                Tdew.append(value(model.props[1].temperature_dew["Vap", "Liq"]))

            elif hasattr(model.props[1], "_mole_frac_tbub"):
                print("One of the components only exists in liquid phase.")
                Tbubb.append(value(model.props[1].temperature_bubble["Vap", "Liq"]))

            X.append(v)

        # If the solver did not solve to an optimal solution, do not store the data point
        else:
            print(f"Case: {count} No Result {component_1} x = {x_d[i]:.2f}")
        count += 1

    # Call TXYData function and store the data in TD class
    TD = TXYDataClass(component_1, component_2, Punit, Tunit, pressure)
    TD.TBubb = Tbubb
    TD.TDew = Tdew
    TD.x = X

    # Return the data class with all the information of the calculations
    return TD


# Author: Alejandro Garciadiego
class TXYDataClass:
    """
    Write data needed for build_txy_diagrams() into a class. The class can be
    obtained by running Txy_data() or by assigining values to the class.
    """

    def __init__(self, component_1, component_2, Punits, Tunits, pressure):
        """
        Args:
            component_1: Component 1
            component_2: Component 2
            Punits: Initial value of heat of hot utility
            Tunits: Initial value of heat to be removed by cold utility
            pressure: Pressure at which the T-x-y data was evaluated

        Returns:
            (Class): A class containing the T-x-y data
        """

        # Build
        self.Component_1 = component_1
        self.Component_2 = component_2

        # Assign units of pressure and temperature
        self.Punits = Punits
        self.Tunits = Tunits

        # Assign pressure at which the data has been calculated
        self.P = pressure

        # Create arrays for data
        self.TBubb = []
        self.TDew = []
        self.x = []

    def Temp_Bubb(self, data_list):
        """
        Args:
            data_list: Bubble temperature array

        Returns:
            None
        """
        self.TBubb = data_list

    def Temp_Dew(self, data_list_2):
        """
        Args:
            data_list_2: Dew temperature array

        Returns:
            None
        """
        self.Tdew = data_list_2

    def composition(self, data_list_3):
        """
        Args:
            data_list_3: x data array

        Returns:
            None
        """
        self.x = data_list_3


# Author: Alejandro Garciadiego
def build_txy_diagrams(
    txy_data, figure_name=None, print_legend=True, include_pressure=False
):
    """
    Args:
        txy_data: Txy data class includes components bubble and dew
        temperatures, compositions, components, pressure, and units.
        figure_name: if a figure name is included the plot will save with the name
        figure_name.png
        print_legend (bool): = If True, include legend to distinguish between
            Bubble and dew temperature. The default is True.
        include_pressure (bool) = If True, print pressure at which the plot is
        calculated in legends. The default is False.
    Returns:
        t-x-y plot
    """

    # Declare a plot and it's size
    ig, ax = plt.subplots(figsize=(12, 8))  # pylint: disable=unused-variable

    if len(txy_data.TBubb) and len(txy_data.TDew) > 0:
        if include_pressure is True:
            # Plot results for bubble temperature

            ax.plot(
                txy_data.x,
                txy_data.TBubb,
                "r",
                label="Bubble Temp P = "
                + str(txy_data.Press)
                + " "
                + str(txy_data.Punits),
                linewidth=1.5,
            )

            # Plot results for dew temperature
            ax.plot(
                txy_data.x,
                txy_data.TDew,
                "b",
                label="Dew Temp P = "
                + str(txy_data.Press)
                + " "
                + str(txy_data.Punits),
                linewidth=1.5,
            )

        else:
            # Plot results for bubble temperature
            ax.plot(
                txy_data.x, txy_data.TBubb, "r", label="Bubble Temp ", linewidth=1.5
            )

            # Plot results for dew temperature
            ax.plot(txy_data.x, txy_data.TDew, "b", label="Dew Temp", linewidth=1.5)

    elif len(txy_data.TDew) == 0:

        if include_pressure is True:
            # Plot results for bubble temperature

            # Plot results for dew temperature
            ax.plot(
                txy_data.x,
                txy_data.TBubb,
                "b",
                label="Dew Temp P = "
                + str(txy_data.Press)
                + " "
                + str(txy_data.Punits),
                linewidth=1.5,
            )

        else:
            # Plot results for dew temperature
            ax.plot(txy_data.x, txy_data.TBubb, "b", label="Dew Temp", linewidth=1.5)

    elif len(txy_data.TBubb) == 0:

        if include_pressure is True:
            # Plot results for bubble temperature

            # Plot results for dew temperature
            ax.plot(
                txy_data.x,
                txy_data.TDew,
                "b",
                label="Bubble Temp P = "
                + str(txy_data.Press)
                + " "
                + str(txy_data.Punits),
                linewidth=1.5,
            )

        else:
            # Plot results for dew temperature
            ax.plot(txy_data.x, txy_data.TDew, "b", label="Dew Temp", linewidth=1.5)

    # Include grid
    ax.grid()

    # Declare labels and fontsize
    plt.xlabel(txy_data.Component_1 + " concentration (mol/mol)", fontsize=20)
    plt.ylabel("Temperature [" + str(txy_data.Tunits) + "]", fontsize=20)

    # Declare plot title
    plt.title(
        "T-x-y diagram " + txy_data.Component_1 + "-" + txy_data.Component_2,
        fontsize=24,
    )

    # Set limits of 0-1 mole fraction
    plt.xlim(0.0, 1)

    # Declare legend and fontsize
    if print_legend is True:
        plt.legend(fontsize=16)

    if figure_name:
        plt.savefig(str(figure_name) + ".png")

    # Show plot
    plt.show()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains classes for reaction blocks and reaction parameter blocks.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, Bool
from pyomo.environ import Var, Constraint, Expression

# Import IDAES cores
from idaes.core.base.process_block import ProcessBlock
from idaes.core import ProcessBlockData, MaterialFlowBasis
from idaes.core.base import property_meta
from idaes.core.util.exceptions import (
    PropertyPackageError,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_reaction_parameter_block,
    is_state_block,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.base.util import build_on_demand
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
)

from idaes.core.util import scaling as iscale
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

# Some more information about this module
__author__ = "Andrew Lee, John Eslick"

__all__ = [
    "ReactionBlock",  # pylint: disable=undefined-all-variable
    "ReactionParameterBlock",  # pylint: disable=undefined-all-variable
]


class _lock_attribute_creation_context(object):
    """Context manager to lock creation of new attributes on a state block"""

    def __init__(self, block):
        self.block = block

    def __enter__(self):
        self.block._lock_attribute_creation = True

    def __exit__(self, exc_type, exc_value, traceback):
        self.block._lock_attribute_creation = False


class ReactionParameterBlock(ProcessBlockData, property_meta.HasPropertyClassMetadata):
    """
    This is the base class for reaction parameter blocks. These are blocks
    that contain a set of parameters associated with a specific reaction
    package, and are linked to by all instances of that reaction package.
    """

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "property_package",
        ConfigValue(
            description="Reference to associated PropertyPackageParameter object",
            domain=is_physical_parameter_block,
        ),
    )
    CONFIG.declare(
        "default_arguments",
        ConfigBlock(
            description="Default arguments to use with Property Package", implicit=True
        ),
    )

    def build(self):
        """
        General build method for ReactionParameterBlocks. Inheriting models
        should call super().build.

        Args:
            None

        Returns:
            None
        """
        super(ReactionParameterBlock, self).build()

        if not hasattr(self, "_reaction_block_class"):
            self._reaction_block_class = None

        # TODO: Need way to tie reaction package to a specific property package
        self._validate_property_parameter_units()
        self._validate_property_parameter_properties()

    @property
    def reaction_block_class(self):
        if self._reaction_block_class is not None:
            return self._reaction_block_class
        else:
            raise AttributeError(
                "{} has not assigned a ReactionBlock class to be associated "
                "with this reaction package. Please contact the developer of "
                "the reaction package.".format(self.name)
            )

    def build_reaction_block(self, *args, **kwargs):
        """
        Methods to construct a ReactionBlock associated with this
        ReactionParameterBlock. This will automatically set the parameters
        construction argument for the ReactionBlock.

        Returns:
            ReactionBlock

        """
        default = kwargs.pop("default", {})
        initialize = kwargs.pop("initialize", {})

        if initialize == {}:
            default["parameters"] = self
        else:
            for i in initialize.keys():
                initialize[i]["parameters"] = self

        return self.reaction_block_class(  # pylint: disable=not-callable
            *args, **kwargs, **default, initialize=initialize
        )

    def _validate_property_parameter_units(self):
        """
        Checks that the property parameter block associated with the
        reaction block uses the same set of default units.
        """
        r_units = self.get_metadata().default_units
        prop_units = self.config.property_package.get_metadata().default_units
        if not r_units.unitset_is_consistent(prop_units):
            raise PropertyPackageError(
                "{} the property package associated with this "
                "reaction package does not use the same set of "
                "units of measurement. Please choose a "
                "property package which uses the same units.".format(self.name)
            )

    def _validate_property_parameter_properties(self):
        """
        Checks that the property parameter block associated with the
        reaction block supports the necessary properties with correct units.
        """
        unsupported = self.get_metadata().properties.check_required_properties(
            self.config.property_package.get_metadata().properties
        )

        if len(unsupported) > 0:
            raise PropertyPackageError(
                f"{self.name} the property package associated with this "
                "reaction package does not support the following necessary "
                "properties, {unsupported}. Please choose a property package "
                "which supports all required properties."
            )


class ReactionBlockBase(ProcessBlock):
    """
    This is the base class for reaction block objects. These are used when
    constructing the SimpleBlock or IndexedBlock which will contain the
    PropertyData objects, and contains methods that can be applied to
    multiple ReactionBlockData objects simultaneously.
    """

    # Set default initializer
    default_initializer = BlockTriangularizationInitializer

    def initialize(self, *args):
        """
        This is a default initialization routine for ReactionBlocks to ensure
        that a routine is present. All ReactionBlockData classes should
        overload this method with one suited to the particular reaction package

        Args:
            None

        Returns:
            None
        """
        raise NotImplementedError(
            "{} reaction package has not implemented the"
            " initialize method. Please contact "
            "the reaction package developer".format(self.name)
        )

    def report(self, index=(0), true_state=False, dof=False, ostream=None, prefix=""):
        raise NotImplementedError(
            """The current Reaction Package has not implemented a report
                method. Please contact the package developer about this."""
        )


class ReactionBlockDataBase(ProcessBlockData):
    """
    This is the base class for reaction block data objects. These are
    blocks that contain the Pyomo components associated with calculating a
    set of reacion properties for a given material.
    """

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "parameters",
        ConfigValue(
            domain=is_reaction_parameter_block,
            description="""A reference to an instance of the Reaction Parameter
Block associated with this property package.""",
        ),
    )
    CONFIG.declare(
        "state_block",
        ConfigValue(
            domain=is_state_block,
            description="""A reference to an instance of a StateBlock with
which this reaction block should be associated.""",
        ),
    )
    CONFIG.declare(
        "has_equilibrium",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Equilibrium constraint flag",
            doc="""Flag indicating whether equilibrium constraints
should be constructed in this reaction block,
**default** - True.
**Valid values:** {
**True** - ReactionBlock should enforce equilibrium constraints,
**False** - ReactionBlock should not enforce equilibrium constraints.}""",
        ),
    )

    def __init__(self, *args, **kwargs):
        self._lock_attribute_creation = False
        super().__init__(*args, **kwargs)

    @property
    def component_list(self):
        return self.state_ref.component_list

    @property
    def phase_list(self):
        return self.state_ref.phase_list

    @property
    def phase_component_set(self):
        return self.state_ref.phase_component_set

    @property
    def default_initializer(self):
        return self.parent_component().default_initializer

    @property
    def default_scaler(self):
        return self.parent_component().default_scaler

    def lock_attribute_creation_context(self):
        """Returns a context manager that does not allow attributes to be created
        while in the context and allows attributes to be created normally outside
        the context.
        """
        return _lock_attribute_creation_context(self)

    def is_property_constructed(self, attr):
        """Returns True if the attribute ``attr`` already exists, or false if it
        would be added in ``__getattr__``, or does not exist.

        Args:
            attr (str): Attribute name to check

        Return:
            True if the attribute is already constructed, False otherwise
        """
        with self.lock_attribute_creation_context():
            return hasattr(self, attr)

    def build(self):
        """
        General build method for PropertyBlockDatas. Inheriting models should
        call super().build.

        Args:
            None

        Returns:
            None
        """
        super(ReactionBlockDataBase, self).build()
        add_object_reference(self, "_params", self.config.parameters)

        self._validate_state_block()

    @property
    def params(self):
        return self._params

    def _validate_state_block(self):
        """
        Method to validate that the associated state block matches with the
        PropertyParameterBlock associated with the ReactionParameterBlock.
        """
        # Add a reference to the corresponding state block data for later use
        add_object_reference(self, "state_ref", self.config.state_block[self.index()])

        # Validate that property package of state matches that of reaction pack
        if (
            self.config.parameters.config.property_package
            != self.state_ref.config.parameters
        ):
            raise PropertyPackageError(
                "{} the StateBlock associated with this "
                "ReactionBlock does not match with the "
                "PropertyParameterBlock associated with the "
                "ReactionParameterBlock. The modelling framework "
                "does not support mixed associations of property "
                "and reaction packages.".format(self.name)
            )

    def get_reaction_rate_basis(self):
        """
        Method which returns an Enum indicating the basis of the reaction rate
        term.
        """
        return MaterialFlowBasis.other

    def __getattr__(self, attr):
        """
        This method is used to avoid generating unnecessary property
        calculations in reaction blocks. __getattr__ is called whenever a
        property is called for, and if a property does not exist, it looks for
        a method to create the required property, and any associated
        components.

        Create a property calculation if needed. Return an attribute error if
        attr == 'domain' or starts with a _ . The error for _ prevents a
        recursion error if trying to get a function to create a property and
        that function doesn't exist.  Pyomo also occasionally looks for things
        that start with _ and may not exist.  Pyomo also looks for the domain
        attribute, and it may not exist.
        This works by creating a property calculation by calling the "_"+attr
        function.

        A list of __getattr__ calls is maintained in self.__getattrcalls to
        check for recursive loops which maybe useful for debugging. This list
        is cleared after __getattr__ completes successfully.

        Args:
            attr: an attribute to create and return. Should be a property
                  component.
        """
        try:
            # try the Pyomo Block's __getattr__ method first which will return
            # decorators for creating components on the block (e.g. Expression,
            # Constraint, ...).
            return super().__getattr__(attr)
        except AttributeError:
            return build_on_demand(self, attr)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # Get scaling factor defaults, if no scaling factor set
        for v in self.component_data_objects(
            (Constraint, Var, Expression), descend_into=False
        ):
            if iscale.get_scaling_factor(v) is None:  # don't replace if set
                name = v.getname().split("[")[0]
                index = v.index()
                sf = self.config.parameters.get_default_scaling(name, index)
                if sf is not None:
                    iscale.set_scaling_factor(v, sf)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import os
import pytest
import idaes
import logging


class TestIdaesConfigure(object):
    @pytest.mark.unit
    def test_write_config(self):
        idaes.write_config(
            os.path.join(idaes.testing_directory, "config_testing_orig.json")
        )
        idaes.write_config(
            os.path.join(idaes.testing_directory, "config_testing_default.json")
        )

    @pytest.mark.unit
    def test_read_default_config(self):
        # set some stuff to not the default
        idaes.cfg.use_idaes_solvers = False
        assert idaes.cfg.use_idaes_solvers == False
        # read the default
        idaes.read_config(
            os.path.join(idaes.testing_directory, "config_testing_default.json")
        )
        assert idaes.cfg.ipopt.options.nlp_scaling_method == "gradient-based"
        assert idaes.cfg.use_idaes_solvers == True

    @pytest.mark.unit
    def test_change_env(self):
        with idaes.temporary_config_ctx():
            # default config uses idaes solvers
            idaes.cfg.use_idaes_solvers = False
            idaes.reconfig()
            pf = os.environ["PATH"]
        pt = os.environ["PATH"]
        assert pf != pt


@pytest.mark.unit
def test_canonical_arch():
    assert idaes.config.canonical_arch("Intel64") == "x86_64"
    assert idaes.config.canonical_arch("AMD64") == "x86_64"
    assert idaes.config.canonical_arch("ARM64") == "aarch64"


@pytest.mark.unit
def test_canonical_distro():
    assert idaes.config.canonical_distro("kubUntu1804") == "ubuntu1804"
    assert idaes.config.canonical_distro("Ubuntu1804") == "ubuntu1804"
    assert idaes.config.canonical_distro("Windows") == "windows"
    assert idaes.config.canonical_distro("daRwin") == "darwin"


@pytest.mark.unit
def test_warning_to_except():
    with idaes.temporary_config_ctx():
        with pytest.raises(RuntimeError):
            _log = logging.getLogger("idaes")
            idaes.cfg.warning_to_exception = True
            idaes.reconfig()
            _log.warning("Hey! Don't do that.")


@pytest.mark.unit
def test_deprecate_to_except():
    with idaes.temporary_config_ctx():
        with pytest.raises(RuntimeError):
            _log = logging.getLogger("idaes")
            idaes.cfg.deprecation_to_exception = True
            idaes.reconfig()
            _log.warning("DEPRECATED: Hey! Don't use that.")


@pytest.mark.unit
def test_get_data_directory():
    # this tests the function that give that data, binary and testing directories
    # it doesn't actually change the idaes configuration.
    if "IDAES_DATA" in os.environ:
        odat = os.environ["IDAES_DATA"]
    else:
        odat = None
    # Test set by env variable on a directory that should exist
    os.environ["IDAES_DATA"] = idaes.config.data_directory
    dd, bd, td = idaes.config.get_data_directory()
    assert dd == idaes.config.data_directory
    # Test set by env variable on a directory that should exist
    os.environ["IDAES_DATA"] = "/this_directory_doesnt_exist/for_real/not_here"
    dd, bd, td = idaes.config.get_data_directory()
    assert dd is None
    if odat is not None:
        os.environ["IDAES_DATA"] = odat
    else:
        del os.environ["IDAES_DATA"]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Standard IDAES Feed block with phase equilibrium.
"""
from enum import Enum

# Import Pyomo libraries
from pyomo.environ import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.initialization import fix_state_vars

__author__ = "Andrew Lee"


# Enumerate options for material balances
class FlashType(Enum):
    """
    Enum of supported flash types.
    """

    isothermal = 1
    isenthalpic = 2


@declare_process_block_class("FeedFlash")
class FeedFlashData(UnitModelBlockData):
    """
    Standard Feed block with phase equilibrium
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Feed units do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Feed units do not have defined volume, thus this must be
False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "flash_type",
        ConfigValue(
            default=FlashType.isothermal,
            domain=In(FlashType),
            description="Type of flash to perform",
            doc="""Indicates what type of flash operation should be used.
**default** - FlashType.isothermal.
**Valid values:** {
**FlashType.isothermal** - specify temperature,
**FlashType.isenthalpic** - specify enthalpy.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model.

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(FeedFlashData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        # No need for control volume geometry

        self.control_volume.add_state_blocks(has_phase_equilibrium=True)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type, has_phase_equilibrium=True
        )

        # Add isothermal constraint
        if self.config.flash_type == FlashType.isothermal:

            @self.Constraint(self.flowsheet().time, doc="Isothermal constraint")
            def isothermal(b, t):
                return (
                    b.control_volume.properties_in[t].temperature
                    == b.control_volume.properties_out[t].temperature
                )

        elif self.config.flash_type == FlashType.isenthalpic:

            @self.Constraint(self.flowsheet().time, doc="Isothermal constraint")
            def isenthalpic(b, t):
                cv = b.control_volume
                return sum(
                    cv.properties_in[t].get_enthalpy_flow_terms(p)
                    for p in cv.properties_in[t].phase_list
                ) == sum(
                    cv.properties_out[t].get_enthalpy_flow_terms(p)
                    for p in cv.properties_in[t].phase_list
                )

        self.control_volume.add_momentum_balances(
            balance_type=MomentumBalanceType.pressureTotal
        )

        # Add references to all feed state vars
        s_vars = self.control_volume.properties_in[
            self.flowsheet().time.first()
        ].define_state_vars()
        for s in s_vars:
            l_name = s_vars[s].local_name
            if s_vars[s].is_indexed():
                slicer = self.control_volume.properties_in[:].component(l_name)[...]
            else:
                slicer = self.control_volume.properties_in[:].component(l_name)

            r = Reference(slicer)
            setattr(self, s, r)

        # Add Ports
        self.add_outlet_port()

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {"Outlet": self.outlet}, time_point=time_point
        )

    def fix_initialization_states(self):
        """
        Fix feed state variables

        Returns:
            None
        """
        fix_state_vars(self.control_volume.properties_in)
